import tempfile
from pathlib import Path, PurePosixPath

import pytest
import torch
from fsspec.implementations.http import HTTPFileSystem
from fsspec.implementations.local import LocalFileSystem
from gcsfs import GCSFileSystem
from kedro.io.core import PROTOCOL_DELIMITER, DatasetError, Version
from s3fs.core import S3FileSystem

from kedro_datasets_experimental.safetensors import SafetensorsDataset


@pytest.fixture
def filepath(tmp_path):
    return (tmp_path / "test.safetensors").as_posix()


@pytest.fixture(params=["torch"])
def backend(request):
    return request.param


@pytest.fixture
def safetensors_dataset(filepath, backend, fs_args):
    return SafetensorsDataset(
        filepath=filepath,
        backend=backend,
        fs_args=fs_args,
    )


@pytest.fixture
def versioned_safetensors_dataset(filepath, load_version, save_version):
    return SafetensorsDataset(
        filepath=filepath, version=Version(load_version, save_version)
    )


@pytest.fixture
def dummy_data():
    return {"embeddings": torch.zeros((10, 100))}


class TestSafetensorsDataset:
    @pytest.mark.skip(reason="Skipping for now, need to investigate")
    @pytest.mark.parametrize(
        "backend",
        [
            "torch",
        ],
        indirect=True,
    )
    def test_save_and_load(self, safetensors_dataset, dummy_data):
        """Test saving and reloading the dataset."""
        safetensors_dataset.save(dummy_data)
        reloaded = safetensors_dataset.load()

        if safetensors_dataset._backend == "torch":
            assert torch.equal(dummy_data["embeddings"], reloaded["embeddings"])

        assert safetensors_dataset._fs_open_args_load == {}
        assert safetensors_dataset._fs_open_args_save == {"mode": "wb"}

    @pytest.mark.skip(reason="Skipping for now, need to investigate")
    def test_exists(self, safetensors_dataset, dummy_data):
        """Test `exists` method invocation for both existing and
        nonexistent dataset."""
        assert not safetensors_dataset.exists()
        safetensors_dataset.save(dummy_data)
        assert safetensors_dataset.exists()

    @pytest.mark.parametrize(
        "fs_args",
        [{"open_args_load": {"mode": "rb", "compression": "gzip"}}],
        indirect=True,
    )
    def test_open_extra_args(self, safetensors_dataset, fs_args):
        assert safetensors_dataset._fs_open_args_load == fs_args["open_args_load"]
        assert safetensors_dataset._fs_open_args_save == {"mode": "wb"}  # default unchanged

    def test_load_missing_file(self, safetensors_dataset):
        """Check the error when trying to load missing file."""
        pattern = r"Failed while loading data from dataset kedro_datasets_experimental.safetensors.safetensors_dataset.SafetensorsDataset\(.*\)"
        with pytest.raises(DatasetError, match=pattern):
            safetensors_dataset.load()

    @pytest.mark.parametrize(
        "filepath,instance_type",
        [
            ("s3://bucket/file.safetensors", S3FileSystem),
            (tempfile.NamedTemporaryFile(suffix=".safetensors").name, LocalFileSystem),
            (tempfile.NamedTemporaryFile(suffix=".safetensors").name, LocalFileSystem),
            ("gcs://bucket/file.safetensors", GCSFileSystem),
            ("https://example.com/file.safetensors", HTTPFileSystem),
        ],
    )
    def test_protocol_usage(self, filepath, instance_type):
        dataset = SafetensorsDataset(filepath=filepath)
        assert isinstance(dataset._fs, instance_type)

        path = filepath.split(PROTOCOL_DELIMITER, 1)[-1]

        assert str(dataset._filepath) == path
        assert isinstance(dataset._filepath, PurePosixPath)

    def test_catalog_release(self, mocker):
        fs_mock = mocker.patch("fsspec.filesystem").return_value
        filepath = "test.safetensors"
        dataset = SafetensorsDataset(filepath=filepath)
        dataset.release()
        fs_mock.invalidate_cache.assert_called_once_with(filepath)

    def test_invalid_backend(self, mocker):
        pattern = (
            r"Selected backend 'invalid' could not be imported. "
            r"Make sure it is installed and importable."
        )
        mocker.patch(
            "kedro_datasets_experimental.safetensors.safetensors_dataset.importlib.import_module",
            side_effect=ImportError,
        )
        with pytest.raises(ImportError, match=pattern):
            SafetensorsDataset(filepath="test.safetensors", backend="invalid")

    def test_copy(self, safetensors_dataset):
        safetensors_dataset_copy = safetensors_dataset._copy()
        assert safetensors_dataset_copy is not safetensors_dataset
        assert safetensors_dataset_copy._describe() == safetensors_dataset._describe()


class TestSafetensorsDatasetVersioned:
    def test_version_str_repr(self, load_version, save_version):
        """Test that version is in string representation of the class instance
        when applicable."""
        filepath = "test.safetensors"
        ds = SafetensorsDataset(filepath=filepath)
        ds_versioned = SafetensorsDataset(
            filepath=filepath, version=Version(load_version, save_version)
        )
        assert filepath in str(ds)
        assert "version" not in str(ds)

        assert filepath in str(ds_versioned)
        ver_str = f"version=Version(load={load_version}, save='{save_version}')"
        assert ver_str in str(ds_versioned)
        assert "SafetensorsDataset" in str(ds_versioned)
        assert "SafetensorsDataset" in str(ds)
        assert "protocol" in str(ds_versioned)
        assert "protocol" in str(ds)
        assert "backend" in str(ds_versioned)
        assert "backend" in str(ds)

    @pytest.mark.skip(reason="Skipping for now, need to investigate")
    def test_save_and_load(self, versioned_safetensors_dataset, dummy_data):
        """Test that saved and reloaded data matches the original one for
        the versioned dataset."""
        versioned_safetensors_dataset.save(dummy_data)
        reloaded_df = versioned_safetensors_dataset.load()

        assert torch.equal(dummy_data["embeddings"], reloaded_df["embeddings"])

    def test_no_versions(self, versioned_safetensors_dataset):
        """Check the error if no versions are available for load."""
        pattern = r"Did not find any versions for kedro_datasets_experimental.safetensors.safetensors_dataset.SafetensorsDataset\(.+\)"
        with pytest.raises(DatasetError, match=pattern):
            versioned_safetensors_dataset.load()

    @pytest.mark.skip(reason="Skipping for now, need to investigate")
    def test_exists(self, versioned_safetensors_dataset, dummy_data):
        """Test `exists` method invocation for versioned dataset."""
        assert not versioned_safetensors_dataset.exists()
        versioned_safetensors_dataset.save(dummy_data)
        assert versioned_safetensors_dataset.exists()

    @pytest.mark.skip(reason="Skipping for now, need to investigate")
    def test_prevent_overwrite(self, versioned_safetensors_dataset, dummy_data):
        """Check the error when attempting to override the dataset if the
        corresponding Safetensors file for a given save version already exists."""
        versioned_safetensors_dataset.save(dummy_data)
        pattern = (
            r"Save path \'.+\' for SafetensorsDataset\(.+\) must "
            r"not exist if versioning is enabled\."
        )
        with pytest.raises(DatasetError, match=pattern):
            versioned_safetensors_dataset.save(dummy_data)

    @pytest.mark.skip(reason="Skipping for now, need to investigate")
    @pytest.mark.parametrize(
        "load_version", ["2019-01-01T23.59.59.999Z"], indirect=True
    )
    @pytest.mark.parametrize(
        "save_version", ["2019-01-02T00.00.00.000Z"], indirect=True
    )
    def test_save_version_warning(
        self, versioned_safetensors_dataset, load_version, save_version, dummy_data
    ):
        """Check the warning when saving to the path that differs from
        the subsequent load path."""
        pattern = (
            rf"Save version '{save_version}' did not match load version "
            rf"'{load_version}' for SafetensorsDataset\(.+\)"
        )
        with pytest.warns(UserWarning, match=pattern):
            versioned_safetensors_dataset.save(dummy_data)

    def test_http_filesystem_no_versioning(self):
        pattern = "Versioning is not supported for HTTP protocols."

        with pytest.raises(DatasetError, match=pattern):
            SafetensorsDataset(
                filepath="https://example.com/file.safetensors", version=Version(None, None)
            )

    @pytest.mark.skip(reason="Skipping for now, need to investigate")
    def test_versioning_existing_dataset(
        self, safetensors_dataset, versioned_safetensors_dataset, dummy_data
    ):
        """Check the error when attempting to save a versioned dataset on top of an
        already existing (non-versioned) dataset."""
        safetensors_dataset.save(dummy_data)
        assert safetensors_dataset.exists()
        assert safetensors_dataset._filepath == versioned_safetensors_dataset._filepath
        pattern = (
            f"(?=.*file with the same name already exists in the directory)"
            f"(?=.*{versioned_safetensors_dataset._filepath.parent.as_posix()})"
        )
        with pytest.raises(DatasetError, match=pattern):
            versioned_safetensors_dataset.save(dummy_data)

        # Remove non-versioned dataset and try again
        Path(safetensors_dataset._filepath.as_posix()).unlink()
        versioned_safetensors_dataset.save(dummy_data)
        assert versioned_safetensors_dataset.exists()

    def test_copy(self, versioned_safetensors_dataset):
        safetensors_dataset_copy = versioned_safetensors_dataset._copy()
        assert safetensors_dataset_copy is not versioned_safetensors_dataset
        assert safetensors_dataset_copy._describe() == versioned_safetensors_dataset._describe()



"""Provides interface to Hugging Face transformers and datasets."""

from typing import Any

import lazy_loader as lazy

try:
    from .hugging_face_dataset import HFDataset
except (ImportError, RuntimeError):
    # For documentation builds that might fail due to dependency issues
    # https://github.com/pylint-dev/pylint/issues/4300#issuecomment-1043601901
    HFDataset: Any

try:
    from .transformer_pipeline_dataset import HFTransformerPipelineDataset
except (ImportError, RuntimeError):
    # For documentation builds that might fail due to dependency issues
    # https://github.com/pylint-dev/pylint/issues/4300#issuecomment-1043601901
    HFTransformerPipelineDataset: Any

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={
        "hugging_face_dataset": ["HFDataset"],
        "transformer_pipeline_dataset": ["HFTransformerPipelineDataset"],
    },
)



import shutil
import tempfile

import numpy as np
import pandas as pd
import pytest
from darts import TimeSeries
from darts.models import RNNModel

from kedro_datasets_experimental.darts import DartsTorchModelDataset


@pytest.fixture
def tfm_kwargs():
    return {
        "pl_trainer_kwargs": {
            "accelerator": "cpu",
            "enable_progress_bar": False,
            "enable_model_summary": False,
        }
    }


@pytest.fixture
def tmpdir_checkpoints_for_module():
    """Sets up a temporary directory for checkpoints that will be deleted after the test module (script) finished."""
    temp_work_dir = tempfile.mkdtemp(prefix="darts")
    yield temp_work_dir
    shutil.rmtree(temp_work_dir)


@pytest.fixture
def filepath_model(tmp_path):
    return (tmp_path / "model.pt").as_posix()


@pytest.fixture
def model_class():
    return "RNNModel"


@pytest.fixture
def model_name():
    return "unittest-model-lstm"


@pytest.fixture
def darts_torch_dataset(filepath_model, model_class):
    return DartsTorchModelDataset(
        filepath=filepath_model,
        model_class=model_class,
        load_args={},
        save_args={},
    )


@pytest.fixture
def load_from_checkpoint_darts_torch_dataset(filepath_model, model_class, model_name, tmpdir_checkpoints_for_module):
    return DartsTorchModelDataset(
        filepath=filepath_model,
        model_class=model_class,
        load_args={"load_method": "load_from_checkpoint",
                   "model_name": model_name, "best": False,
                   "map_location": "cpu", "work_dir": tmpdir_checkpoints_for_module},
        save_args={"save_model": False},
    )


@pytest.fixture
def trained_darts_torch_model(filepath_model, model_class, tfm_kwargs, tmpdir_checkpoints_for_module, model_name):
    times = pd.date_range("20130101", "20130410")
    pd_series = pd.Series(range(100), index=times)
    series: TimeSeries = TimeSeries.from_series(pd_series)
    model = RNNModel(
        input_chunk_length=1,
        model="LSTM",
        n_epochs=1,
        model_name=model_name,
        work_dir=tmpdir_checkpoints_for_module,
        save_checkpoints=True,
        force_reset=True,
        **tfm_kwargs,
    )
    model.fit(series)
    return model


class TestDartsTorchModelDataset:
    def test_save_and_load_dataset(self, trained_darts_torch_model, darts_torch_dataset):
        darts_torch_dataset.save(trained_darts_torch_model)

        model_loaded = darts_torch_dataset.load()

        pred1 = trained_darts_torch_model.predict(n=6)
        pred2 = model_loaded.predict(n=6)

        # Two models with the same parameters should deterministically yield the same output
        np.testing.assert_array_equal(pred1.values(), pred2.values())

    def test_load_from_checkpoint_dataset(
            self,
            trained_darts_torch_model,
            tmpdir_checkpoints_for_module,
            load_from_checkpoint_darts_torch_dataset
    ):
        model_loaded = load_from_checkpoint_darts_torch_dataset.load()
        pred1 = trained_darts_torch_model.predict(n=6)
        pred2 = model_loaded.predict(n=6)

        # Two models with the same parameters should deterministically yield the same output
        np.testing.assert_array_equal(pred1.values(), pred2.values())



"""``ExcelDataset`` loads/saves data from/to a Excel file using an underlying
filesystem (e.g.: local, S3, GCS). It uses pandas to handle the Excel file.
"""
from __future__ import annotations

import logging
from copy import deepcopy
from pathlib import PurePosixPath
from typing import Any

import fsspec
import pandas as pd
from kedro.io.core import (
    PROTOCOL_DELIMITER,
    AbstractVersionedDataset,
    DatasetError,
    Version,
    get_filepath_str,
    get_protocol_and_path,
)

from kedro_datasets._typing import TablePreview

logger = logging.getLogger(__name__)


class ExcelDataset(
    AbstractVersionedDataset[
        pd.DataFrame | dict[str, pd.DataFrame],
        pd.DataFrame | dict[str, pd.DataFrame],
    ]
):
    """``ExcelDataset`` loads/saves data from/to a Excel file using an underlying
    filesystem (e.g.: local, S3, GCS). It uses pandas to handle the Excel file.

    To save a multi-sheet Excel file, no special ``save_args`` are required.
    Instead, return a dictionary of ``Dict[str, pd.DataFrame]`` where the string
    keys are your sheet names.

    Examples:
        Using the [YAML API](https://docs.kedro.org/en/stable/catalog-data/data_catalog_yaml_examples/):

        ```yaml
        rockets:
          type: pandas.ExcelDataset
          filepath: gcs://your_bucket/rockets.xlsx
          fs_args:
            project: my-project
          credentials: my_gcp_credentials
          save_args:
            sheet_name: Sheet1
          load_args:
            sheet_name: Sheet1

        shuttles:
          type: pandas.ExcelDataset
          filepath: data/01_raw/shuttles.xlsx
        ```

        Using the [Python API](https://docs.kedro.org/en/stable/catalog-data/advanced_data_catalog_usage/):

        >>> import pandas as pd
        >>> from kedro_datasets.pandas import ExcelDataset
        >>>
        >>> data = pd.DataFrame({"col1": [1, 2], "col2": [4, 5], "col3": [5, 6]})
        >>>
        >>> dataset = ExcelDataset(filepath=tmp_path / "test.xlsx")
        >>> dataset.save(data)
        >>> reloaded = dataset.load()
        >>> assert data.equals(reloaded)

        For a multi-sheet Excel file, using the [YAML API](https://docs.kedro.org/en/stable/catalog-data/data_catalog_yaml_examples/):

        ```yaml
        trains:
          type: pandas.ExcelDataset
          filepath: data/02_intermediate/company/trains.xlsx
          load_args:
            sheet_name: [Sheet1, Sheet2, Sheet3]
        ```

        For a multi-sheet Excel file, using the [Python API](https://docs.kedro.org/en/stable/catalog-data/advanced_data_catalog_usage/):

        >>> import pandas as pd
        >>> from kedro_datasets.pandas import ExcelDataset
        >>>
        >>> dataframe = pd.DataFrame({"col1": [1, 2], "col2": [4, 5], "col3": [5, 6]})
        >>> another_dataframe = pd.DataFrame({"x": [10, 20], "y": ["hello", "world"]})
        >>> multiframe = {"Sheet1": dataframe, "Sheet2": another_dataframe}
        >>>
        >>> dataset = ExcelDataset(filepath="test.xlsx", load_args={"sheet_name": None})
        >>> dataset.save(multiframe)
        >>> reloaded = dataset.load()
        >>> assert multiframe["Sheet1"].equals(reloaded["Sheet1"])
        >>> assert multiframe["Sheet2"].equals(reloaded["Sheet2"])
    """

    DEFAULT_LOAD_ARGS = {"engine": "openpyxl"}
    DEFAULT_SAVE_ARGS = {"index": False}
    DEFAULT_FS_ARGS: dict[str, Any] = {"open_args_save": {"mode": "wb"}}

    def __init__(  # noqa: PLR0913
        self,
        *,
        filepath: str,
        engine: str = "openpyxl",
        load_args: dict[str, Any] | None = None,
        save_args: dict[str, Any] | None = None,
        version: Version | None = None,
        credentials: dict[str, Any] | None = None,
        fs_args: dict[str, Any] | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        """Creates a new instance of ``ExcelDataset`` pointing to a concrete Excel file
        on a specific filesystem.

        Args:
            filepath: Filepath in POSIX format to a Excel file prefixed with a protocol like
                `s3://`. If prefix is not provided, `file` protocol (local filesystem) will be used.
                The prefix should be any protocol supported by ``fsspec``.
                Note: `http(s)` doesn't support versioning.
            engine: The engine used to write to Excel files. The default
                engine is 'openpyxl'.
            load_args: Pandas options for loading Excel files.
                Here you can find all available arguments:
                https://pandas.pydata.org/pandas-docs/stable/generated/pandas.read_excel.html
                All defaults are preserved, but "engine", which is set to "openpyxl".
                Supports multi-sheet Excel files (include `sheet_name = None` in `load_args`).
            save_args: Pandas options for saving Excel files.
                Here you can find all available arguments:
                https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.to_excel.html
                All defaults are preserved, but "index", which is set to False.
                If you would like to specify options for the `ExcelWriter`,
                you can include them under the "writer" key. Here you can
                find all available arguments:
                https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.ExcelWriter.html
            version: If specified, should be an instance of
                ``kedro.io.core.Version``. If its ``load`` attribute is
                None, the latest version will be loaded. If its ``save``
                attribute is None, save version will be autogenerated.
            credentials: Credentials required to get access to the underlying filesystem.
                E.g. for ``GCSFileSystem`` it should look like `{"token": None}`.
            fs_args: Extra arguments to pass into underlying filesystem class constructor
                (e.g. `{"project": "my-project"}` for ``GCSFileSystem``).
                Defaults are preserved, apart from the `open_args_save` `mode` which is set to `wb`.
                Note that the save method requires bytes, so any save mode provided should include "b" for bytes.
            metadata: Any arbitrary metadata.
                This is ignored by Kedro, but may be consumed by users or external plugins.

        Raises:
            DatasetError: If versioning is enabled while in append mode.
        """
        _fs_args = deepcopy(fs_args) or {}
        _fs_open_args_load = _fs_args.pop("open_args_load", {})
        _fs_open_args_save = _fs_args.pop("open_args_save", {})
        _credentials = deepcopy(credentials) or {}

        protocol, path = get_protocol_and_path(filepath, version)
        if protocol == "file":
            _fs_args.setdefault("auto_mkdir", True)

        self._protocol = protocol
        self._storage_options = {**_credentials, **_fs_args}
        self._fs = fsspec.filesystem(self._protocol, **self._storage_options)

        self.metadata = metadata

        super().__init__(
            filepath=PurePosixPath(path),
            version=version,
            exists_function=self._fs.exists,
            glob_function=self._fs.glob,
        )

        # Handle default load and save and fs arguments
        self._load_args = {**self.DEFAULT_LOAD_ARGS, **(load_args or {})}
        self._save_args = {**self.DEFAULT_SAVE_ARGS, **(save_args or {})}
        self._fs_open_args_load = {
            **self.DEFAULT_FS_ARGS.get("open_args_load", {}),
            **(_fs_open_args_load or {}),
        }
        self._fs_open_args_save = {
            **self.DEFAULT_FS_ARGS.get("open_args_save", {}),
            **(_fs_open_args_save or {}),
        }

        self._writer_args = self._save_args.pop("writer", {})  # type: ignore
        self._writer_args.setdefault("engine", engine or "openpyxl")  # type: ignore

        if version and self._writer_args.get("mode") == "a":  # type: ignore
            raise DatasetError(
                "'ExcelDataset' doesn't support versioning in append mode."
            )

        if "storage_options" in self._save_args or "storage_options" in self._load_args:
            logger.warning(
                "Dropping 'storage_options' for %s, "
                "please specify them under 'fs_args' or 'credentials'.",
                self._filepath,
            )
            self._save_args.pop("storage_options", None)
            self._load_args.pop("storage_options", None)

    def _describe(self) -> dict[str, Any]:
        return {
            "filepath": self._filepath,
            "protocol": self._protocol,
            "load_args": self._load_args,
            "save_args": self._save_args,
            "writer_args": self._writer_args,
            "version": self._version,
        }

    def load(self) -> pd.DataFrame | dict[str, pd.DataFrame]:
        load_path = str(self._get_load_path())
        if self._protocol == "file":
            # file:// protocol seems to misbehave on Windows
            # (<urlopen error file not on local host>),
            # so we don't join that back to the filepath;
            # storage_options also don't work with local paths
            return pd.read_excel(load_path, **self._load_args)

        load_path = f"{self._protocol}{PROTOCOL_DELIMITER}{load_path}"
        return pd.read_excel(
            load_path, storage_options=self._storage_options, **self._load_args
        )

    def save(self, data: pd.DataFrame | dict[str, pd.DataFrame]) -> None:
        save_path = get_filepath_str(self._get_save_path(), self._protocol)

        with self._fs.open(save_path, **self._fs_open_args_save) as fs_file:
            with pd.ExcelWriter(fs_file, **self._writer_args) as writer:
                if isinstance(data, dict):
                    for sheet_name, sheet_data in data.items():
                        sheet_data.to_excel(
                            writer, sheet_name=sheet_name, **self._save_args
                        )
                else:
                    data.to_excel(writer, **self._save_args)

        self._invalidate_cache()

    def _exists(self) -> bool:
        try:
            load_path = get_filepath_str(self._get_load_path(), self._protocol)
        except DatasetError:
            return False

        return self._fs.exists(load_path)

    def _release(self) -> None:
        super()._release()
        self._invalidate_cache()

    def _invalidate_cache(self) -> None:
        """Invalidate underlying filesystem caches."""
        filepath = get_filepath_str(self._filepath, self._protocol)
        self._fs.invalidate_cache(filepath)

    def preview(self, nrows: int = 5) -> TablePreview:
        """
        Generate a preview of the dataset with a specified number of rows.

        Args:
            nrows: The number of rows to include in the preview. Defaults to 5.

        Returns:
            dict: A dictionary containing the data in a split format.
        """
        # Create a copy so it doesn't contaminate the original dataset
        dataset_copy = self._copy()
        dataset_copy._load_args["nrows"] = nrows  # type: ignore[attr-defined]
        data = dataset_copy.load()

        return data.to_dict(orient="split")



import uuid

import chromadb
import pytest
from kedro.io.core import DatasetError

from kedro_datasets_experimental.chromadb import ChromaDBDataset


@pytest.fixture
def sample_data():
    """Sample data for testing ChromaDB operations."""
    return {
        "documents": [
            "This is the first document about machine learning",
            "This is the second document about data science",
            "This is the third document about artificial intelligence"
        ],
        "ids": ["doc1", "doc2", "doc3"],
        "metadatas": [
            {"topic": "ml", "author": "Alice"},
            {"topic": "ds", "author": "Bob"},
            {"topic": "ai", "author": "Charlie"}
        ]
    }


@pytest.fixture
def chromadb_dataset():
    """ChromaDB dataset with ephemeral client for testing."""
    # Use unique collection name to avoid test interference
    unique_name = f"test_collection_{uuid.uuid4().hex[:8]}"
    return ChromaDBDataset(
        collection_name=unique_name,
        client_type="ephemeral"
    )


@pytest.fixture
def persistent_chromadb_dataset(tmp_path):
    """ChromaDB dataset with persistent client for testing."""
    # Use unique collection name to avoid test interference
    unique_name = f"test_collection_{uuid.uuid4().hex[:8]}"
    return ChromaDBDataset(
        collection_name=unique_name,
        client_type="persistent",
        client_settings={"path": str(tmp_path / "test_chroma_db")}
    )


class TestChromaDBDataset:
    """Test suite for ChromaDBDataset."""

    def test_save_and_load(self, chromadb_dataset, sample_data):
        """Test saving and loading data."""
        # Save data
        chromadb_dataset.save(sample_data)

        # Load data back
        loaded_data = chromadb_dataset.load()

        # Verify the data
        assert loaded_data["documents"] == sample_data["documents"]
        assert loaded_data["ids"] == sample_data["ids"]
        assert loaded_data["metadatas"] == sample_data["metadatas"]

    def test_exists(self, chromadb_dataset, sample_data):
        """Test the exists method."""
        # Initially, collection shouldn't exist or be empty
        assert not chromadb_dataset.exists()

        # After saving data, it should exist
        chromadb_dataset.save(sample_data)
        assert chromadb_dataset.exists()

    def test_save_without_required_fields(self, chromadb_dataset):
        """Test saving data without required fields raises error."""
        # Missing 'ids'
        invalid_data = {"documents": ["test doc"]}
        with pytest.raises(DatasetError, match="Data must contain 'documents' and 'ids' keys"):
            chromadb_dataset.save(invalid_data)

        # Missing 'documents'
        invalid_data = {"ids": ["doc1"]}
        with pytest.raises(DatasetError, match="Data must contain 'documents' and 'ids' keys"):
            chromadb_dataset.save(invalid_data)

    def test_save_non_dict_data(self, chromadb_dataset):
        """Test saving non-dictionary data raises error."""
        with pytest.raises(DatasetError, match="Data must be a dictionary"):
            chromadb_dataset.save(["not", "a", "dict"])

    @pytest.mark.skip(reason="Skipping for now, need to investigate")
    def test_vector_similarity_query(self, chromadb_dataset, sample_data):
        """Test vector similarity queries without loading entire collection."""
        # Save initial data
        chromadb_dataset.save(sample_data)

        # Create a new dataset instance with query args for similarity search
        query_dataset = ChromaDBDataset(
            collection_name=chromadb_dataset._collection_name,
            client_type="ephemeral",
            load_args={
                "query_texts": ["machine learning"],
                "n_results": 2,
                "include": ["documents", "metadatas", "distances"]
            }
        )

        # Load with vector query - should return limited results based on similarity
        results = query_dataset.load()

        # Should return limited results (not the entire collection)
        assert len(results["documents"]) <= 2
        assert "distances" in results  # Query results include distances

        # Verify it's actually a similarity search result
        assert isinstance(results["documents"], list)
        if results["documents"]:
            # Should contain documents similar to "machine learning"
            assert any("machine" in doc.lower() or "learning" in doc.lower()
                      for doc in results["documents"])

    def test_load_with_query_args(self, chromadb_dataset, sample_data):
        """Test loading data with query arguments."""
        # Save initial data
        chromadb_dataset.save(sample_data)

        # Use the same dataset instance to load with limit
        # Create a new dataset instance with load args using same collection name
        dataset_with_query = ChromaDBDataset(
            collection_name=chromadb_dataset._collection_name,
            client_type="ephemeral",
            load_args={"limit": 2}  # Use limit instead of n_results for get() method
        )

        # Load with query - should return limited results
        loaded_data = dataset_with_query.load()
        assert len(loaded_data["documents"]) <= 2

    def test_save_with_embeddings(self, chromadb_dataset):
        """Test saving data with custom embeddings."""
        # First save some data to establish the embedding dimension
        initial_data = {
            "documents": ["Initial document"],
            "ids": ["initial_doc"]
        }
        chromadb_dataset.save(initial_data)

        # Now try with custom embeddings that match the established dimension
        # ChromaDB uses 384-dim embeddings by default, so skip this test
        # or use a collection with compatible embedding function
        data_with_embeddings = {
            "documents": ["Test document"],
            "ids": ["doc1"],
            # Skip custom embeddings as they need to match the embedding function dimension
        }

        # Should not raise an error
        chromadb_dataset.save(data_with_embeddings)

        # Verify data was saved
        loaded_data = chromadb_dataset.load()
        assert "Test document" in loaded_data["documents"]
        assert "doc1" in loaded_data["ids"]

    def test_persistent_client(self, persistent_chromadb_dataset, sample_data):
        """Test ChromaDB with persistent client."""
        # Save data
        persistent_chromadb_dataset.save(sample_data)

        # Load data back
        loaded_data = persistent_chromadb_dataset.load()

        # Verify the data
        assert loaded_data["documents"] == sample_data["documents"]
        assert loaded_data["ids"] == sample_data["ids"]

    def test_describe(self, chromadb_dataset):
        """Test the describe method."""
        description = chromadb_dataset._describe()

        assert description["collection_name"] == chromadb_dataset._collection_name
        assert description["client_type"] == "ephemeral"
        assert isinstance(description["client_settings"], dict)
        assert isinstance(description["load_args"], dict)
        assert isinstance(description["save_args"], dict)

    def test_invalid_client_type(self):
        """Test invalid client type raises error."""
        dataset = ChromaDBDataset(
            collection_name="test",
            client_type="invalid"
        )
        # Error should be raised when we try to use the dataset (lazy loading)
        with pytest.raises(DatasetError, match="Unsupported client_type: invalid"):
            dataset._get_client()

    def test_http_client_config(self):
        """Test HTTP client configuration without connecting."""
        # This test creates the dataset but doesn't trigger client creation
        # since we don't have a real HTTP server running
        dataset = ChromaDBDataset(
            collection_name="test_collection",
            client_type="http",
            client_settings={"host": "localhost", "port": 8000}
        )

        # Test the configuration is stored correctly
        description = dataset._describe()
        assert description["client_type"] == "http"
        assert description["client_settings"]["host"] == "localhost"
        assert description["client_settings"]["port"] == 8000

        # Test that the client is not created yet (lazy initialization)
        assert dataset._client is None

    def test_metadata_handling(self, chromadb_dataset, sample_data):
        """Test that metadata is properly handled."""
        # Create dataset with metadata
        dataset_with_metadata = ChromaDBDataset(
            collection_name="test_collection",
            client_type="ephemeral",
            metadata={"version": "1.0", "description": "Test dataset"}
        )

        # Metadata should be stored
        assert dataset_with_metadata.metadata == {"version": "1.0", "description": "Test dataset"}

        # Normal operations should still work
        dataset_with_metadata.save(sample_data)
        loaded_data = dataset_with_metadata.load()
        assert loaded_data["documents"] == sample_data["documents"]

    def test_collection_reuse(self, sample_data):
        """Test that multiple dataset instances can use the same collection."""
        # Use a unique collection name for this test
        collection_name = f"shared_collection_{uuid.uuid4().hex[:8]}"

        # Create two dataset instances pointing to the same collection
        dataset1 = ChromaDBDataset(collection_name=collection_name, client_type="ephemeral")
        dataset2 = ChromaDBDataset(collection_name=collection_name, client_type="ephemeral")

        # Save data with first dataset
        dataset1.save(sample_data)

        # For ephemeral clients, different instances don't share collections
        # This test verifies the pattern works for persistent clients
        # With ephemeral clients, this is expected behavior
        try:
            loaded_data = dataset2.load()
            # If it works, verify the data (this would work with persistent clients)
            assert loaded_data["documents"] == sample_data["documents"]
        except Exception:
            # Ephemeral clients don't share collections, which is expected
            # This test demonstrates the difference between client types
            pass

    def test_empty_collection_load(self, chromadb_dataset):
        """Test loading from an empty collection."""
        # First create an empty collection by accessing it
        chromadb_dataset._get_collection()

        # Load from empty collection should return empty lists
        loaded_data = chromadb_dataset.load()

        assert loaded_data["documents"] == []
        assert loaded_data["ids"] == []
        assert loaded_data["metadatas"] == []
        # Check embeddings length instead of direct comparison (avoids numpy array comparison issue)
        assert len(loaded_data["embeddings"]) == 0



"""Provide data loading and saving functionality for Ibis's backends."""
from typing import Any

import lazy_loader as lazy

try:
    from .file_dataset import FileDataset
except (ImportError, RuntimeError):
    # For documentation builds that might fail due to dependency issues
    # https://github.com/pylint-dev/pylint/issues/4300#issuecomment-1043601901
    FileDataset: Any

try:
    from .table_dataset import TableDataset
except (ImportError, RuntimeError):
    # For documentation builds that might fail due to dependency issues
    # https://github.com/pylint-dev/pylint/issues/4300#issuecomment-1043601901
    TableDataset: Any

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={"file_dataset": ["FileDataset"], "table_dataset": ["TableDataset"]},
)



"""Provides I/O for TensorFlow models."""

from typing import Any

import lazy_loader as lazy

try:
    from .tensorflow_model_dataset import TensorFlowModelDataset
except (ImportError, RuntimeError):
    TensorFlowModelDataset: Any

# https://github.com/pylint-dev/pylint/issues/4300#issuecomment-1043601901
__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={"tensorflow_model_dataset": ["TensorFlowModelDataset"]},
)



import os

import boto3
import pytest
import xarray as xr
from kedro.io.core import DatasetError
from moto import mock_aws
from s3fs import S3FileSystem
from xarray.testing import assert_equal

from kedro_datasets_experimental.netcdf import NetCDFDataset

FILE_NAME = "test.nc"
MULTIFILE_NAME = "test*.nc"
BUCKET_NAME = "test_bucket"
MULTIFILE_BUCKET_NAME = "test_bucket_multi"
AWS_CREDENTIALS = {"key": "FAKE_ACCESS_KEY", "secret": "FAKE_SECRET_KEY"}

# Pathlib cannot be used since it strips out the second slash from "s3://"
S3_PATH = f"s3://{BUCKET_NAME}/{FILE_NAME}"
S3_PATH_MULTIFILE = f"s3://{MULTIFILE_BUCKET_NAME}/{MULTIFILE_NAME}"


@pytest.fixture
def mocked_s3_bucket():
    """Create a bucket for testing to store a singular NetCDF file."""
    with mock_aws():
        conn = boto3.client(
            "s3",
            aws_access_key_id=AWS_CREDENTIALS["key"],
            aws_secret_access_key=AWS_CREDENTIALS["secret"],
        )
        conn.create_bucket(Bucket=BUCKET_NAME)
        yield conn


@pytest.fixture
def mocked_s3_bucket_multi():
    """Create a bucket for testing to store multiple NetCDF files."""
    with mock_aws():
        conn = boto3.client(
            "s3",
            aws_access_key_id=AWS_CREDENTIALS["key"],
            aws_secret_access_key=AWS_CREDENTIALS["secret"],
        )
        conn.create_bucket(Bucket=MULTIFILE_BUCKET_NAME)
        yield conn


def dummy_data() -> xr.Dataset:
    """Sample xarray dataset for load/save testing."""
    ds = xr.DataArray(
        [0, 1, 2, 3], dims=["x"], coords={"x": [0, 1, 2, 3]}, name="data"
    ).to_dataset()
    return ds


@pytest.fixture
def dummy_xr_dataset() -> xr.Dataset:
    """Expected result for load/save on a single NetCDF file."""
    return dummy_data()


@pytest.fixture
def dummy_xr_dataset_multi() -> xr.Dataset:
    """Expected concatenated result for load/save on multiple NetCDF files."""
    data = dummy_data()
    return xr.concat([data, data], dim="dummy")


@pytest.fixture
def mocked_s3_object(tmp_path, mocked_s3_bucket, dummy_xr_dataset: xr.Dataset):
    """Creates singular test NetCDF and adds it to mocked S3 bucket."""
    temporary_path = tmp_path / FILE_NAME
    dummy_xr_dataset.to_netcdf(str(temporary_path))

    mocked_s3_bucket.put_object(
        Bucket=BUCKET_NAME, Key=FILE_NAME, Body=temporary_path.read_bytes()
    )
    return mocked_s3_bucket


@pytest.fixture
def mocked_s3_object_multi(
    tmp_path, mocked_s3_bucket_multi, dummy_xr_dataset: xr.Dataset
):
    """Creates multiple test NetCDFs and adds them to mocked S3 bucket."""

    def put_data(file_name: str):
        temporary_path = tmp_path / file_name
        dummy_xr_dataset.to_netcdf(str(temporary_path))
        mocked_s3_bucket_multi.put_object(
            Bucket=MULTIFILE_BUCKET_NAME,
            Key=file_name,
            Body=temporary_path.read_bytes(),
        )
        return mocked_s3_bucket_multi

    mocked_s3_bucket_multi = put_data("test1.nc")
    mocked_s3_bucket_multi = put_data("test2.nc")
    return mocked_s3_bucket_multi


@pytest.fixture
def s3_dataset(load_args, save_args, tmp_path):
    """Sample NetCDF dataset pointing to mocked S3 bucket with single NetCDF file."""
    return NetCDFDataset(
        filepath=S3_PATH,
        temppath=tmp_path,
        credentials=AWS_CREDENTIALS,
        load_args=load_args,
        save_args=save_args,
    )


@pytest.fixture
def s3_dataset_multi(save_args, tmp_path):
    """Sample NetCDF dataset pointing to mocked S3 bucket with multiple NetCDF files."""
    return NetCDFDataset(
        filepath=S3_PATH_MULTIFILE,
        temppath=tmp_path,
        credentials=AWS_CREDENTIALS,
        load_args={"concat_dim": "dummy", "combine": "nested"},
        save_args=save_args,
    )


@pytest.fixture()
def s3fs_cleanup():
    # clear cache so we get a clean slate every time we instantiate a S3FileSystem
    yield
    S3FileSystem.cachable = False


@pytest.mark.usefixtures("s3fs_cleanup")
class TestNetCDFDataset:
    os.environ["AWS_ACCESS_KEY_ID"] = "FAKE_ACCESS_KEY"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "FAKE_SECRET_KEY"

    def test_temppath_error_raised(self):
        """Test that error is raised if S3 NetCDF file referenced without a temporary
        path."""
        pattern = "Need to set temppath in catalog"
        with pytest.raises(ValueError, match=pattern):
            NetCDFDataset(
                filepath=S3_PATH,
                temppath=None,
            )

    @pytest.mark.parametrize("bad_credentials", [{"key": None, "secret": None}])
    def test_empty_credentials_load(self, bad_credentials, tmp_path):
        """Test that error is raised if there are no AWS credentials."""
        netcdf_dataset = NetCDFDataset(
            filepath=S3_PATH, temppath=tmp_path, credentials=bad_credentials
        )
        pattern = r"Failed while loading data from dataset kedro_datasets_experimental.netcdf.netcdf_dataset.NetCDFDataset\(.+\)"
        with pytest.raises(DatasetError, match=pattern):
            netcdf_dataset.load()

    @pytest.mark.xfail(reason="Pending rewrite with new s3fs version")
    def test_pass_credentials(self, mocker, tmp_path):
        """Test that AWS credentials are passed successfully into boto3
        client instantiation on creating S3 connection."""
        client_mock = mocker.patch("botocore.session.Session.create_client")
        s3_dataset = NetCDFDataset(
            filepath=S3_PATH, temppath=tmp_path, credentials=AWS_CREDENTIALS
        )
        pattern = r"Failed while loading data from dataset kedro_datasets_experimental.netcdf.netcdf_dataset.NetCDFDataset\(.+\)"
        with pytest.raises(DatasetError, match=pattern):
            s3_dataset.load()

        assert client_mock.call_count == 1
        args, kwargs = client_mock.call_args_list[0]
        assert args == ("s3",)
        assert kwargs["aws_access_key_id"] == AWS_CREDENTIALS["key"]
        assert kwargs["aws_secret_access_key"] == AWS_CREDENTIALS["secret"]

    @pytest.mark.skip(reason="S3 tests that load datasets don't work properly")
    def test_save_data_single(self, s3_dataset, dummy_xr_dataset, mocked_s3_bucket):
        """Test saving a single NetCDF file to S3."""
        s3_dataset.save(dummy_xr_dataset)
        loaded_data = s3_dataset.load()
        assert_equal(loaded_data, dummy_xr_dataset)

    def test_save_data_multi_error(self, s3_dataset_multi, dummy_xr_dataset_multi):
        """Test that error is raised when trying to save to a NetCDF destination with
        a glob pattern."""
        pattern = r"Globbed multifile datasets with '*'"
        with pytest.raises(DatasetError, match=pattern):
            s3_dataset_multi.save(dummy_xr_dataset)

    @pytest.mark.skip(reason="S3 tests that load datasets don't work properly")
    def test_load_data_single(self, s3_dataset, dummy_xr_dataset, mocked_s3_object):
        """Test loading a single NetCDF file from S3."""
        loaded_data = s3_dataset.load()
        assert_equal(loaded_data, dummy_xr_dataset)

    @pytest.mark.skip(reason="S3 tests that load datasets don't work properly")
    def test_load_data_multi(
        self, s3_dataset_multi, dummy_xr_dataset_multi, mocked_s3_object_multi
    ):
        """Test loading multiple NetCDF files from S3."""
        loaded_data = s3_dataset_multi.load()
        assert_equal(loaded_data, dummy_xr_dataset_multi)

    @pytest.mark.skip(reason="Skipping for now, need to investigate")
    def test_exists(self, s3_dataset, dummy_xr_dataset, mocked_s3_bucket):
        """Test `exists` method invocation for both existing and nonexistent single
        NetCDF file."""
        assert not s3_dataset.exists()
        s3_dataset.save(dummy_xr_dataset)
        assert s3_dataset.exists()

    @pytest.mark.skip(reason="Skipping for now, need to investigate")
    @pytest.mark.usefixtures("mocked_s3_object_multi")
    def test_exists_multi_remote(self, s3_dataset_multi):
        """Test `exists` method invocation works for multifile glob pattern on S3."""
        assert s3_dataset_multi.exists()

    def test_exists_multi_locally(self, tmp_path, dummy_xr_dataset):
        """Test `exists` method invocation for both existing and nonexistent set of
        multiple local NetCDF files."""
        dataset = NetCDFDataset(filepath=str(tmp_path / MULTIFILE_NAME))
        assert not dataset.exists()
        NetCDFDataset(filepath=str(tmp_path / "test1.nc")).save(dummy_xr_dataset)
        NetCDFDataset(filepath=str(tmp_path / "test2.nc")).save(dummy_xr_dataset)
        assert dataset.exists()

    @pytest.mark.parametrize(
        "save_args, load_args",
        [
            ({"engine": "netcdf4"}, {"engine": "netcdf4"}),
            ({"engine": "scipy"}, {"engine": "scipy"}),
            ({"engine": "h5netcdf"}, {"engine": "h5netcdf"}),
        ],
        indirect=True,
    )
    def test_save_load_locally(self, tmp_path, dummy_xr_dataset, save_args, load_args):
        """Test loading and saving the a NetCDF file locally."""
        file_path = str(tmp_path / FILE_NAME)
        dataset = NetCDFDataset(filepath=file_path)

        assert not dataset.exists()
        dataset.save(dummy_xr_dataset)
        assert dataset.exists()
        loaded_data = dataset.load()
        dummy_xr_dataset.equals(loaded_data)

    def test_load_locally_multi(
        self, tmp_path, dummy_xr_dataset, dummy_xr_dataset_multi
    ):
        """Test loading multiple NetCDF files locally."""
        file_path = str(tmp_path / MULTIFILE_NAME)
        dataset = NetCDFDataset(
            filepath=file_path, load_args={"concat_dim": "dummy", "combine": "nested"}
        )

        assert not dataset.exists()
        NetCDFDataset(filepath=str(tmp_path / "test1.nc")).save(dummy_xr_dataset)
        NetCDFDataset(filepath=str(tmp_path / "test2.nc")).save(dummy_xr_dataset)
        assert dataset.exists()
        loaded_data = dataset.load()
        dummy_xr_dataset_multi.equals(loaded_data.compute())

    @pytest.mark.parametrize(
        "load_args", [{"k1": "v1", "index": "value"}], indirect=True
    )
    def test_load_extra_params(self, s3_dataset, load_args):
        """Test overriding the default load arguments."""
        for key, value in load_args.items():
            assert s3_dataset._load_args[key] == value

    @pytest.mark.parametrize(
        "save_args", [{"k1": "v1", "index": "value"}], indirect=True
    )
    def test_save_extra_params(self, s3_dataset, save_args):
        """Test overriding the default save arguments."""
        for key, value in save_args.items():
            assert s3_dataset._save_args[key] == value

        for key, value in s3_dataset.DEFAULT_SAVE_ARGS.items():
            assert s3_dataset._save_args[key] == value



import json
from pathlib import Path
from typing import Any

import pytest
import yaml
from kedro.io import DatasetError

try:
    # LangChain >= 1.0
    from langchain_core.prompts import ChatPromptTemplate, PromptTemplate
except ImportError:
    # LangChain < 1.0
    from langchain.prompts import ChatPromptTemplate, PromptTemplate

from kedro_datasets_experimental.langchain.langchain_prompt_dataset import (
    LangChainPromptDataset,
)


@pytest.fixture
def txt_prompt_file(tmp_path: Path) -> Path:
    """Create a temporary text file with a simple prompt."""
    prompt_file = tmp_path / "prompt.txt"
    prompt_file.write_text("Hello, this is {name}!")
    return prompt_file


@pytest.fixture
def json_prompt_file(tmp_path: Path) -> Path:
    """Create a temporary JSON file with a prompt configuration."""
    prompt_file = tmp_path / "prompt.json"
    prompt_file.write_text(
        """
        {
            "template": "Hello, this is {name}!",
            "input_variables": ["name"]
        }
        """
    )
    return prompt_file


@pytest.fixture
def chat_prompt_file(tmp_path: Path) -> Path:
    """Create a temporary JSON file with a chat prompt configuration."""
    prompt_file = tmp_path / "chat.json"
    prompt_file.write_text(
        """
        {
            "messages": [
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": "Hello, this is {name}!"}
            ]
        }
        """
    )
    return prompt_file


@pytest.fixture
def yaml_prompt_file(tmp_path: Path) -> Path:
    """Create a temporary YAML file with a simple prompt configuration."""
    prompt_file = tmp_path / "prompt.yaml"
    content = {
        "template": "Hello, this is {name}!",
        "input_variables": ["name"]
    }
    prompt_file.write_text(yaml.dump(content))
    return prompt_file


@pytest.fixture
def chat_yaml_prompt_file(tmp_path: Path) -> Path:
    """Create a temporary YAML file with a chat prompt configuration."""
    prompt_file = tmp_path / "chat_prompt.yaml"
    content = {
        "messages": [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "Hello, this is {name}!"}
        ]
    }
    prompt_file.write_text(yaml.dump(content))
    return prompt_file


class TestLangChainPromptDataset:
    def test_init_with_txt_file(self, txt_prompt_file: Path) -> None:
        """Test dataset initialization with a .txt file."""
        dataset = LangChainPromptDataset(
            filepath=str(txt_prompt_file),
            dataset={"type": "text.TextDataset"}
            )
        assert dataset._template_name == "PromptTemplate"
        assert dataset._dataset.__class__.__name__ == "TextDataset"

    def test_init_with_json_file(self, json_prompt_file: Path) -> None:
        """Test dataset initialization with a .json file."""
        dataset = LangChainPromptDataset(
            filepath=str(json_prompt_file),
            dataset={"type": "json.JSONDataset"}
            )
        assert dataset._template_name == "PromptTemplate"
        assert dataset._dataset.__class__.__name__ == "JSONDataset"

    def test_init_with_invalid_template(self, txt_prompt_file: Path) -> None:
        """Test initialization with invalid template type."""
        with pytest.raises(DatasetError, match="Invalid template"):
            LangChainPromptDataset(
                filepath=str(txt_prompt_file),
                dataset={"type": "text.TextDataset"},
                template="InvalidTemplate"
            )

    def test_load_text_prompt(self, txt_prompt_file: Path) -> None:
        """Test loading a simple text prompt."""
        dataset = LangChainPromptDataset(
            filepath=str(txt_prompt_file),
            dataset={"type": "kedro_datasets.text.TextDataset"}
            )
        prompt = dataset.load()
        assert isinstance(prompt, PromptTemplate)
        assert prompt.format(name="Kedro") == "Hello, this is Kedro!"

    def test_load_json_prompt(self, json_prompt_file: Path) -> None:
        """Test loading a JSON prompt configuration."""
        dataset = LangChainPromptDataset(
            filepath=str(json_prompt_file),
            dataset={"type": "kedro_datasets.json.JSONDataset"})
        prompt = dataset.load()
        assert isinstance(prompt, PromptTemplate)
        assert prompt.format(name="Kedro") == "Hello, this is Kedro!"

    def test_load_chat_prompt(self, chat_prompt_file: Path) -> None:
        """Test loading a chat prompt configuration."""
        dataset = LangChainPromptDataset(
            filepath=str(chat_prompt_file),
            dataset={"type": "json.JSONDataset"},
            template="ChatPromptTemplate"
        )
        prompt = dataset.load()
        assert isinstance(prompt, ChatPromptTemplate)
        messages = prompt.format_messages(name="Kedro")
        assert len(messages) == 2
        assert messages[0].content == "You are a helpful assistant."
        assert messages[1].content == "Hello, this is Kedro!"

    def test_load_yaml_prompt(self, yaml_prompt_file: Path) -> None:
        """Test loading a YAML prompt configuration."""
        dataset = LangChainPromptDataset(
            filepath=str(yaml_prompt_file),
            dataset={"type": "kedro_datasets.yaml.YAMLDataset"}
            )
        prompt = dataset.load()
        assert isinstance(prompt, PromptTemplate)
        assert prompt.format(name="Kedro") == "Hello, this is Kedro!"

    def test_load_yaml_chat_prompt(self, chat_yaml_prompt_file: Path) -> None:
        """Test loading a YAML chat prompt configuration."""
        dataset = LangChainPromptDataset(
            filepath=str(chat_yaml_prompt_file),
            dataset={"type": "yaml.YAMLDataset"},
            template="ChatPromptTemplate"
        )
        prompt = dataset.load()
        assert isinstance(prompt, ChatPromptTemplate)
        messages = prompt.format_messages(name="Kedro")
        assert len(messages) == 2
        assert messages[0].content == "You are a helpful assistant."
        assert messages[1].content == "Hello, this is Kedro!"


    def test_save_not_supported(self, txt_prompt_file: Path) -> None:
        """Test that save operation raises an error."""
        dataset = LangChainPromptDataset(
            filepath=str(txt_prompt_file),
            dataset={"type": "text.TextDataset"}
            )
        with pytest.raises(DatasetError, match="Saving is not supported"):
            dataset.save({})

    def test_describe(self, txt_prompt_file: Path) -> None:
        """Test the _describe method returns correct information."""
        dataset = LangChainPromptDataset(
            filepath=str(txt_prompt_file),
            dataset={"type": "text.TextDataset"})
        desc = dataset._describe()
        assert desc["template"] == "PromptTemplate"
        assert desc["underlying_dataset"] == "TextDataset"
        assert "path" in desc

    def test_exists(self, txt_prompt_file: Path) -> None:
        """Test the _exists method."""
        dataset = LangChainPromptDataset(
            filepath=str(txt_prompt_file),
            dataset={"type": "text.TextDataset"}
            )
        assert dataset._exists() is True

    def test_credentials_propagation(self, json_prompt_file: Path) -> None:
        """Test that credentials are properly propagated to the underlying dataset."""
        credentials = {"key": "value"}
        dataset = LangChainPromptDataset(
            filepath=str(json_prompt_file),
            dataset={"type": "json.JSONDataset"},
            credentials=credentials,
        )

        # JSONDataset stores credentials in the fsspec filesystem options
        # We are just checking that the credentials were passed correctly
        fs_options = getattr(dataset._dataset._fs, "storage_options", {})
        for k, v in credentials.items():
            assert fs_options.get(k) == v

    def test_chat_prompt_template_with_plain_string(self, tmp_path: Path) -> None:
        """Test that plain string data raises DatasetError for ChatPromptTemplate."""
        prompt_file = tmp_path / "plain_string.json"
        prompt_file.write_text('"Just a plain string, not a chat prompt."')
        dataset = LangChainPromptDataset(
            filepath=str(prompt_file),
            dataset={"type": "text.TextDataset"},
            template="ChatPromptTemplate"
        )
        with pytest.raises(DatasetError, match="Plain string data is only supported for PromptTemplate, not ChatPromptTemplate."):
            dataset.load()

    def test_invalid_dataset_type_raises_error(self, txt_prompt_file: Path) -> None:
        """Test that using an invalid dataset type raises DatasetError."""
        invalid_dataset = {"type": "pandas.CSVDataset"}
        with pytest.raises(DatasetError, match="Unsupported dataset type 'pandas.CSVDataset'"):
            LangChainPromptDataset(filepath=str(txt_prompt_file), dataset=invalid_dataset)

    def test_none_dataset_type_raises_error(self, txt_prompt_file: Path) -> None:
        """Test that passing no dataset type raises DatasetError."""
        invalid_dataset = None
        with pytest.raises(DatasetError, match="Underlying dataset type cannot be empty"):
            LangChainPromptDataset(filepath=str(txt_prompt_file), dataset=invalid_dataset)

    @pytest.mark.parametrize(
        "bad_data,error_pattern",
        [
            ({}, "ChatPromptTemplate requires a non-empty list of messages"),
            ([], "ChatPromptTemplate requires a non-empty list of messages"),
            (("single item"), "Expecting value"),
            ([{"content": "hello"}], "Expecting property name enclosed in double quotes"),
        ],
    )
    def test_invalid_chat_prompt_data(
        self, tmp_path: Path, bad_data: Any, error_pattern: str
    ) -> None:
        """Test validation of chat prompt data."""
        prompt_file = tmp_path / "bad_chat.json"
        prompt_file.write_text(str(bad_data))
        dataset = LangChainPromptDataset(
            filepath=str(prompt_file),
            dataset={"type": "json.JSONDataset"},
            template="ChatPromptTemplate"
        )

        with pytest.raises(DatasetError, match=f".*{error_pattern}.*"):
            dataset.load()

    @pytest.mark.parametrize(
        "bad_data,error_pattern",
        [
            ({}, "ChatPromptTemplate requires a non-empty list of messages"),
            ([], "ChatPromptTemplate requires a non-empty list of messages"),
            ([{"role": "user"}], "Expected dict to have exact keys 'role' and 'content'"),
            ([{"content": "hello"}], "Expected dict to have exact keys 'role' and 'content'"),
        ]
    )
    def test_invalid_yaml_chat_prompt(self, tmp_path: Path, bad_data: any, error_pattern: str) -> None:
        """Test invalid YAML chat prompts for various error cases."""
        prompt_file = tmp_path / "bad_chat.yaml"
        prompt_file.write_text(yaml.dump(bad_data))

        dataset = LangChainPromptDataset(
            filepath=str(prompt_file),
            dataset={"type": "yaml.YAMLDataset"},
            template="ChatPromptTemplate"
        )

        with pytest.raises(DatasetError, match=f".*{error_pattern}.*"):
            dataset.load()

    def test_preview_txt_prompt(self, txt_prompt_file: str):
        """Preview a plain text prompt returns raw string."""
        dataset = LangChainPromptDataset(
            filepath=str(txt_prompt_file),
            dataset={"type": "text.TextDataset"}
            )
        preview = dataset.preview()
        assert preview == '{"text": "Hello, this is {name}!"}'

    def test_preview_json_prompt(self, json_prompt_file: str):
        """Preview a JSON prompt returns serialized dict."""
        dataset = LangChainPromptDataset(
            filepath=str(json_prompt_file),
            dataset={"type": "json.JSONDataset"}
            )
        preview = dataset.preview()
        data = json.loads(preview)
        assert data["template"] == "Hello, this is {name}!"
        assert data["input_variables"] == ["name"]

    def test_preview_yaml_prompt(self, yaml_prompt_file: str):
        """Preview a YAML prompt returns serialized dict."""
        dataset = LangChainPromptDataset(
            filepath=str(yaml_prompt_file),
            dataset={"type": "yaml.YAMLDataset"}
            )
        preview = dataset.preview()
        data = json.loads(preview)
        assert data["template"] == "Hello, this is {name}!"
        assert data["input_variables"] == ["name"]

    def test_preview_chat_json_prompt(self, chat_prompt_file: str):
        """Preview a JSON chat prompt returns serialized messages list."""
        dataset = LangChainPromptDataset(
            filepath=str(chat_prompt_file),
            dataset={"type": "json.JSONDataset"},
            template="ChatPromptTemplate")
        preview = dataset.preview()
        data = json.loads(preview)
        assert isinstance(data, dict)
        assert data['messages'][0]['role'] == 'system'
        assert data['messages'][1]['role'] == 'user'

    def test_preview_chat_yaml_prompt(self, chat_yaml_prompt_file: str):
        """Preview a YAML chat prompt returns serialized messages list."""
        dataset = LangChainPromptDataset(
            filepath=str(chat_yaml_prompt_file),
            dataset={"type": "yaml.YAMLDataset"},
            template="ChatPromptTemplate")
        preview = dataset.preview()
        data = json.loads(preview)
        assert isinstance(data, dict)
        assert data['messages'][0]['role'] == 'system'
        assert data['messages'][1]['role'] == 'user'



"""Provide file loading and saving functionality for Ibis's backends."""
from __future__ import annotations

from copy import deepcopy
from pathlib import Path, PurePosixPath
from typing import TYPE_CHECKING, Any, ClassVar

import ibis.expr.types as ir
from kedro.io import AbstractVersionedDataset, DatasetError, Version

from kedro_datasets._utils import ConnectionMixin

if TYPE_CHECKING:
    from ibis import BaseBackend


class FileDataset(ConnectionMixin, AbstractVersionedDataset[ir.Table, ir.Table]):
    """``FileDataset`` loads/saves data from/to a specified file format.

    Examples:
        Using the [YAML API](https://docs.kedro.org/en/stable/catalog-data/data_catalog_yaml_examples/):

        ```yaml
        cars:
          type: ibis.FileDataset
          filepath: data/01_raw/company/cars.csv
          file_format: csv
          table_name: cars
          connection:
            backend: duckdb
            database: company.db
          load_args:
            sep: ","
            nullstr: "#NA"
          save_args:
            sep: ","
            nullstr: "#NA"

        motorbikes:
          type: ibis.FileDataset
          filepath: s3://your_bucket/data/02_intermediate/company/motorbikes/
          file_format: delta
          table_name: motorbikes
          connection:
            backend: polars
        ```

        Using the [Python API](https://docs.kedro.org/en/stable/catalog-data/advanced_data_catalog_usage/):

        >>> import ibis
        >>> from kedro_datasets.ibis import FileDataset
        >>>
        >>> data = ibis.memtable({"col1": [1, 2], "col2": [4, 5], "col3": [5, 6]})
        >>>
        >>> dataset = FileDataset(
        ...     filepath=tmp_path / "test.csv",
        ...     file_format="csv",
        ...     table_name="test",
        ...     connection={"backend": "duckdb", "database": tmp_path / "file.db"},
        ... )
        >>> dataset.save(data)
        >>> reloaded = dataset.load()
        >>> assert data.execute().equals(reloaded.execute())

    """

    DEFAULT_CONNECTION_CONFIG: ClassVar[dict[str, Any]] = {
        "backend": "duckdb",
        "database": ":memory:",
    }
    DEFAULT_LOAD_ARGS: ClassVar[dict[str, Any]] = {}
    DEFAULT_SAVE_ARGS: ClassVar[dict[str, Any]] = {}

    _CONNECTION_GROUP: ClassVar[str] = "ibis"

    def __init__(  # noqa: PLR0913
        self,
        filepath: str,
        file_format: str = "parquet",
        *,
        table_name: str | None = None,
        connection: dict[str, Any] | None = None,
        load_args: dict[str, Any] | None = None,
        save_args: dict[str, Any] | None = None,
        version: Version | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        """Creates a new ``FileDataset`` pointing to the given filepath.

        ``FileDataset`` connects to the Ibis backend object constructed
        from the connection configuration. The `backend` key provided in
        the config can be any of the `supported backends <https://ibis-\
        project.org/install>`_. The remaining dictionary entries will be
        passed as arguments to the underlying ``connect()`` method (e.g.
        `ibis.duckdb.connect() <https://ibis-project.org/backends/duckdb\
        #ibis.duckdb.connect>`_).

        The read method corresponding to the given ``file_format`` (e.g.
        `read_csv() <https://ibis-project.org/backends/\
        duckdb#ibis.backends.duckdb.Backend.read_csv>`_) is used to load
        the file with the backend. Note that only the data is loaded; no
        link to the underlying file exists past ``FileDataset.load()``.

        Args:
            filepath: Path to a file to register as a table. Most useful
                for loading data into your data warehouse (for testing).
                On save, the backend exports data to the specified path.
            file_format: String specifying the file format for the file.
                Defaults to writing execution results to a Parquet file.
            table_name: The name to use for the created table (on load).
            connection: Configuration for connecting to an Ibis backend.
                If not provided, connect to DuckDB in in-memory mode.
            load_args: Additional arguments passed to the Ibis backend's
                `read_{file_format}` method.
            save_args: Additional arguments passed to the Ibis backend's
                `to_{file_format}` method.
            version: If specified, should be an instance of
                ``kedro.io.core.Version``. If its ``load`` attribute is
                None, the latest version will be loaded. If its ``save``
                attribute is None, save version will be autogenerated.
            metadata: Any arbitrary metadata. This is ignored by Kedro,
                but may be consumed by users or external plugins.
        """
        self._file_format = file_format
        self._table_name = table_name
        self._connection_config = connection or self.DEFAULT_CONNECTION_CONFIG
        self.metadata = metadata

        super().__init__(
            filepath=PurePosixPath(filepath),
            version=version,
            exists_function=lambda filepath: Path(filepath).exists(),
        )

        # Set load and save arguments, overwriting defaults if provided.
        self._load_args = deepcopy(self.DEFAULT_LOAD_ARGS)
        if load_args is not None:
            self._load_args.update(load_args)

        self._save_args = deepcopy(self.DEFAULT_SAVE_ARGS)
        if save_args is not None:
            self._save_args.update(save_args)

    def _connect(self) -> BaseBackend:
        import ibis  # noqa: PLC0415

        config = deepcopy(self._connection_config)
        backend = getattr(ibis, config.pop("backend"))
        return backend.connect(**config)

    @property
    def connection(self) -> BaseBackend:
        """The ``Backend`` instance for the connection configuration."""
        return self._connection

    def load(self) -> ir.Table:
        load_path = self._get_load_path()
        reader = getattr(self.connection, f"read_{self._file_format}")
        return reader(load_path, table_name=self._table_name, **self._load_args)

    def save(self, data: ir.Table) -> None:
        save_path = self._get_save_path()
        Path(save_path).parent.mkdir(parents=True, exist_ok=True)
        writer = getattr(self.connection, f"to_{self._file_format}")
        writer(data, save_path, **self._save_args)

    def _describe(self) -> dict[str, Any]:
        return {
            "filepath": self._filepath,
            "file_format": self._file_format,
            "table_name": self._table_name,
            "backend": self._connection_config["backend"],
            "load_args": self._load_args,
            "save_args": self._save_args,
            "version": self._version,
        }

    def _exists(self) -> bool:
        try:
            load_path = self._get_load_path()
        except DatasetError:
            return False

        return Path(load_path).exists()



"""``DeltaTableDataset`` loads/saves delta tables from/to a filesystem (e.g.: local,
S3, GCS), Databricks unity catalog and AWS Glue catalog respectively. It handles
load and save using a pandas dataframe.
"""

from __future__ import annotations

from copy import deepcopy
from typing import Any

import pandas as pd
from deltalake import DeltaTable, Metadata
from deltalake.exceptions import TableNotFoundError
from deltalake.writer import write_deltalake
from kedro.io.core import AbstractDataset, DatasetError


class DeltaTableDataset(AbstractDataset):
    """``DeltaTableDataset`` loads/saves delta tables from/to a filesystem (e.g.: local,
    S3, GCS), Databricks unity catalog and AWS Glue catalog respectively. It handles
    load and save using a pandas dataframe. When saving data, you can specify one of two
    modes: overwrite(default), append. If you wish to alter the schema as a part of
    overwrite, pass overwrite_schema=True. You can overwrite a specific partition by using
    mode=overwrite together with partition_filters. This will remove all files within the
    matching partition and insert your data as new files.

    Examples:
        Using the [YAML API](https://docs.kedro.org/en/stable/catalog-data/data_catalog_yaml_examples/):

        ```yaml
        boats_filesystem:
          type: pandas.DeltaTableDataset
          filepath: data/01_raw/boats
          credentials: dev_creds
          load_args:
            version: 7
          save_args:
            mode: overwrite

        boats_databricks_unity_catalog:
          type: pandas.DeltaTableDataset
          credentials: dev_creds
          catalog_type: UNITY
          database: simple_database
          table: simple_table
          save_args:
            mode: overwrite

        trucks_aws_glue_catalog:
          type: pandas.DeltaTableDataset
          credentials: dev_creds
          catalog_type: AWS
          catalog_name: main
          database: db_schema
          table: db_table
          save_args:
            mode: overwrite
        ```

        Using the [Python API](https://docs.kedro.org/en/stable/catalog-data/advanced_data_catalog_usage/):

        >>> from kedro_datasets.pandas import DeltaTableDataset
        >>> import pandas as pd
        >>>
        >>> data = pd.DataFrame({"col1": [1, 2], "col2": [4, 5], "col3": [5, 6]})
        >>>
        >>> dataset = DeltaTableDataset(filepath=tmp_path / "test")
        >>> dataset.save(data)
        >>> reloaded = dataset.load()
        >>> assert data.equals(reloaded)
        >>>
        >>> new_data = pd.DataFrame({"col1": [7, 8], "col2": [9, 10], "col3": [11, 12]})
        >>> dataset.save(new_data)
        >>> assert isinstance(dataset.get_loaded_version(), int)

    """

    DEFAULT_WRITE_MODE = "overwrite"
    ACCEPTED_WRITE_MODES = ("overwrite", "append")

    DEFAULT_LOAD_ARGS: dict[str, Any] = {}
    DEFAULT_SAVE_ARGS: dict[str, Any] = {"mode": DEFAULT_WRITE_MODE}

    def __init__(  # noqa: PLR0913
        self,
        *,
        filepath: str | None = None,
        catalog_type: str | None = None,
        catalog_name: str | None = None,
        database: str | None = None,
        table: str | None = None,
        load_args: dict[str, Any] | None = None,
        save_args: dict[str, Any] | None = None,
        credentials: dict[str, Any] | None = None,
        fs_args: dict[str, Any] | None = None,
    ) -> None:
        """Creates a new instance of ``DeltaTableDataset``

        Args:
            filepath (str): Filepath to a delta lake file with the following accepted protocol:
                ``S3``: `s3://<bucket>/<path>`, `s3a://<bucket>/<path>`
                ``Azure``: `az://<container>/<path>`, `adl://<container>/<path>`,
                `abfs://<container>/<path>`
                ``GCS``: `gs://<bucket>/<path>`
                If any of the prefix above is not provided, `file` protocol (local filesystem)
                will be used.
            catalog_type (DataCatalog, Optional): `AWS` or `UNITY` if filepath is not provided.
                Defaults to None.
            catalog_name (str, Optional): the name of catalog in AWS Glue or Databricks Unity.
                Defaults to None.
            database (str, Optional): the name of the database (also referred to as schema).
                Defaults to None.
            table (str, Optional): the name of the table.
            load_args (Dict[str, Any], Optional): Additional options for loading file(s)
                into DeltaTableDataset. `load_args` accepts `version` to load the appropriate
                version when loading from a filesystem.
            save_args (Dict[str, Any], Optional): Additional saving options for saving into
                Delta lake. Here you can find all available arguments:
                https://delta-io.github.io/delta-rs/python/api_reference.html#writing-deltatables
            credentials (Dict[str, Any], Optional): Credentials required to get access to
                the underlying filesystem. E.g. for ``GCSFileSystem`` it should look like
                `{"token": None}`.
            fs_args (Dict[str, Any], Optional): Extra arguments to pass into underlying
                filesystem class constructor.
                (e.g. `{"project": "my-project"}` for ``GCSFileSystem``).
        Raises:
            DatasetError: Invalid configuration supplied (through DeltaTableDataset validation)

        """
        self._filepath = filepath
        self._catalog_type = catalog_type
        self._catalog_name = catalog_name
        self._database = database
        self._table = table
        self._fs_args = deepcopy(fs_args or {})
        self._credentials = deepcopy(credentials or {})

        # DeltaTable cannot be instantiated from an empty directory
        # for the first time creation from filepath, we need to delay the instantiation
        self.is_empty_dir: bool = False
        self._delta_table: DeltaTable | None = None

        # Handle default load and save arguments
        self._load_args = {**self.DEFAULT_LOAD_ARGS, **(load_args or {})}
        self._save_args = {**self.DEFAULT_SAVE_ARGS, **(save_args or {})}

        write_mode = self._save_args.get("mode", None)
        if write_mode not in self.ACCEPTED_WRITE_MODES:
            raise DatasetError(
                f"Write mode {write_mode} is not supported, "
                f"Please use any of the following accepted modes "
                f"{self.ACCEPTED_WRITE_MODES}"
            )

        self._version = self._load_args.get("version", None)

        if self._filepath and self._catalog_type:
            raise DatasetError(
                "DeltaTableDataset can either load from "
                "filepath or catalog_type. Please provide "
                "one of either filepath or catalog_type."
            )

        if self._filepath:
            try:
                self._delta_table = DeltaTable(
                    table_uri=self._filepath,
                    storage_options=self.fs_args,
                    version=self._version,
                )
            except TableNotFoundError:
                self.is_empty_dir = True
        elif self._catalog_type:
            if self._catalog_type.upper() == "AWS":
                table_uri = f"glue:///{self._database}/{self._table}"
            elif self._catalog_type.upper() == "UNITY":
                table_uri = (
                    f"unity://{self._catalog_name}/{self._database}/{self._table}"
                )
            else:
                raise ValueError(f"Unsupported catalog type: {self._catalog_type}")

            self._delta_table = DeltaTable(
                table_uri=table_uri,
                storage_options=self.fs_args,
                version=self._version,
            )

    @property
    def fs_args(self) -> dict[str, Any]:
        """Appends and returns filesystem credentials to fs_args."""
        fs_args = deepcopy(self._fs_args)
        fs_args.update(self._credentials)
        return fs_args

    @property
    def schema(self) -> str:
        """Returns the schema of the DeltaTableDataset as a dictionary."""
        return self._delta_table.schema().to_json() if self._delta_table else ""

    @property
    def metadata(self) -> Metadata | None:
        """Returns the metadata of the DeltaTableDataset as a dictionary.
        Metadata contains the following:
        1. A unique id
        2. A name, if provided
        3. A description, if provided
        4. The list of partition_columns.
        5. The created_time of the table
        6. A map of table configuration. This includes fields such as delta.appendOnly,
        which if true indicates the table is not meant to have data deleted from it.

        Returns: Metadata object containing the above metadata attributes.
        """
        return self._delta_table.metadata() if self._delta_table else None

    @property
    def history(self) -> list[dict[str, Any]] | None:
        """Returns the history of actions on DeltaTableDataset as a list of dictionaries."""
        return self._delta_table.history() if self._delta_table else None

    def get_loaded_version(self) -> int | None:
        """Returns the version of the DeltaTableDataset that is currently loaded."""
        return self._delta_table.version() if self._delta_table else None

    def load(self) -> pd.DataFrame:
        return self._delta_table.to_pandas() if self._delta_table else None

    def save(self, data: pd.DataFrame) -> None:
        if self.is_empty_dir:
            # first time creation of delta table
            write_deltalake(
                self._filepath or "",
                data,
                storage_options=self.fs_args,
                **self._save_args,
            )
            self.is_empty_dir = False
            self._delta_table = DeltaTable(
                table_uri=self._filepath or "",
                storage_options=self.fs_args,
                version=self._version,
            )
        else:
            write_deltalake(
                self._delta_table or "",
                data,
                storage_options=self.fs_args,
                **self._save_args,
            )

    def _describe(self) -> dict[str, Any]:
        return {
            "filepath": self._filepath,
            "catalog_type": self._catalog_type,
            "catalog_name": self._catalog_name,
            "database": self._database,
            "table": self._table,
            "load_args": self._load_args,
            "save_args": self._save_args,
            "version": self._version,
        }



"""``StudyDataset`` loads/saves data from/to an optuna Study."""

from __future__ import annotations

import fnmatch
import logging
import os
from copy import deepcopy
from pathlib import PurePosixPath
from typing import Any

import optuna
from kedro.io.core import (
    AbstractVersionedDataset,
    DatasetError,
    Version,
)
from sqlalchemy import URL
from sqlalchemy.dialects import registry

logger = logging.getLogger(__name__)


class StudyDataset(AbstractVersionedDataset[optuna.Study, optuna.Study]):
    """``StudyDataset`` loads/saves data from/to an optuna Study.

    Example usage for the
    `YAML API <https://docs.kedro.org/en/stable/catalog-data/data_catalog_yaml_examples/>`_:

    .. code-block:: yaml

        review_prediction_study:
          type: kedro_datasets_experimental.optuna.StudyDataset
          backend: sqlite
          database: data/05_model_input/review_prediction_study.db
          load_args:
            sampler:
              class: TPESampler
              n_startup_trials: 10
              n_ei_candidates: 5
            pruner:
              class: NopPruner
          versioned: true

        price_prediction_study:
          type: kedro_datasets_experimental.optuna.StudyDataset
          backend: postgresql
          database: optuna_db
          credentials: dev_optuna_postgresql

    Example usage for the
    `Python API <https://docs.kedro.org/en/stable/data/\
    advanced_data_catalog_usage.html>`_:

    .. code-block:: pycon

        >>> from kedro_datasets_experimental.optuna import StudyDataset
        >>> from optuna.distributions import FloatDistribution
        >>> import optuna
        >>>
        >>> study = optuna.create_study()
        >>> trial = optuna.trial.create_trial(
        ...     params={"x": 2.0},
        ...     distributions={"x": FloatDistribution(0, 10)},
        ...     value=4.0,
        ... )
        >>> study.add_trial(trial)
        >>>
        >>> dataset = StudyDataset(backend="sqlite", database="optuna.db")
        >>> dataset.save(study)
        >>> reloaded = dataset.load()
        >>> assert len(reloaded.trials) == 1
        >>> assert reloaded.trials[0].params["x"] == 2.0
    """

    DEFAULT_LOAD_ARGS: dict[str, Any] = {"sampler": None, "pruner": None}

    def __init__(  # noqa: PLR0913
        self,
        *,
        backend: str,
        database: str,
        study_name: str,
        load_args: dict[str, Any] | None = None,
        version: Version = None,
        credentials: dict[str, Any] | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        """Creates a new instance of ``StudyDataset`` pointing to a concrete optuna
        Study on a specific relational database.

        Args:
            backend: Name of the database backend. This name should correspond to a module
                in ``SQLAlchemy``.
            database: Name of the database.
            study_name: Name of the optuna Study.
            load_args: Optuna options for loading studies. Accepts a `sampler` and a
                `pruner`. If either are provided, a `class` matching any Optuna `sampler`,
                respecitively `pruner` class name should be provided, optionally with
                their argyments. Here you can find all available samplers and pruners
                and their arguments:
                - https://optuna.readthedocs.io/en/stable/reference/samplers/index.html
                - https://optuna.readthedocs.io/en/stable/reference/pruners.html
                All defaults are preserved.
            version: If specified, should be an instance of
                ``kedro.io.core.Version``. If its ``load`` attribute is
                None, the latest version will be loaded. If its ``save``
                attribute is None, save version will be autogenerated.
            credentials: Credentials required to get access to the underlying RDB.
                They can include `username`, `password`, `host`, and `port`.
            metadata: Any arbitrary metadata.
                This is ignored by Kedro, but may be consumed by users or external plugins.
        """
        self._backend = self._validate_backend(backend=backend)
        self._database = self._validate_database(backend=backend, database=database)
        self._study_name = self._validate_study_name(study_name=study_name)

        credentials = self._validate_credentials(backend=backend, credentials=credentials)
        storage_url = URL.create(
            drivername=backend,
            database=database,
            **credentials,
        )

        self._storage_url = storage_url
        self.metadata = metadata

        filepath = None
        if backend == "sqlite":
            filepath = PurePosixPath(os.path.realpath(database))

        super().__init__(
            filepath=filepath,
            version=version,
            exists_function=self._study_name_exists,
            glob_function=self._study_name_glob,
        )

        # Handle default load and save and fs arguments
        self._load_args = {**self.DEFAULT_LOAD_ARGS, **(load_args or {})}

    def _validate_backend(self, backend):
        valid_backends = list(registry.impls.keys()) + ["mssql", "mysql", "oracle", "postgresql", "sqlite"]
        if backend not in valid_backends:
            raise ValueError(
                f"Requested `backend` '{backend}' is not registered as an SQLAlchemy dialect."
            )
        return backend

    def _validate_database(self, backend, database):
        if not isinstance(database, str):
            raise ValueError(f"`database` '{database}' is not a string.")

        if backend == "sqlite":
            if database == ":memory:":
                return database

            # Check if the file has an extension
            _, extension = os.path.splitext(database)
            if not extension:
                raise ValueError(f"The sqlite file `database` '{database}' does not have an extension.")

        return database

    def _validate_study_name(self, study_name):
        if not isinstance(study_name, str):
            raise ValueError(f"`study_name` '{study_name}' is not a string.")
        return study_name

    def _validate_credentials(self, backend, credentials):
        if backend == "sqlite" or credentials is None:
            return {}

        if not set(credentials.keys()) <= {"username", "password", "host", "port"}:
            raise ValueError(
                "Incorrect `credentials`. Provided `credentials` should contain "
                "`'username'`, `'password'`, `'host'`, and/or `'port'`. It contains "
                f"{set(credentials.keys())}."
            )

        return deepcopy(credentials)

    def _get_versioned_path(self, version: str) -> PurePosixPath:
        study_name_posix = PurePosixPath(self._study_name)
        return study_name_posix / version / study_name_posix

    def resolve_load_version(self) -> str | None:
        """Compute the version the dataset should be loaded with."""
        if not self._version:
            return None
        if self._version.load:
            return self._version.load
        return self._fetch_latest_load_version()

    def _get_load_path(self) -> PurePosixPath:
        # Path is not affected by versioning
        return self._filepath

    def _get_load_study_name(self) -> str:
        if not self._version:
            # When versioning is disabled, load from original study name
            return self._study_name

        load_version = self.resolve_load_version()
        return str(self._get_versioned_path(load_version))

    def _get_save_path(self) -> PurePosixPath:
        # Path is not affected by versioning
        return self._filepath

    def _get_save_study_name(self) -> str:
        if not self._version:
            # When versioning is disabled, return original study name
            return self._study_name

        save_version = self.resolve_save_version()
        versioned_study_name = self._get_versioned_path(save_version)

        if self._exists_function(str(versioned_study_name)):
            raise DatasetError(
                f"Study name '{versioned_study_name}' for {self!s} must not exist if "
                f"versioning is enabled."
            )

        return str(versioned_study_name)

    def _describe(self) -> dict[str, Any]:
        return {
            "backend": self._backend,
            "database": self._database,
            "study_name": self._study_name,
            "load_args": self._load_args,
            "version": self._version,
        }

    def _get_sampler(self, sampler_config):
        if sampler_config is None:
            return None

        if "class" not in sampler_config:
            raise ValueError(
                "Optuna `sampler` 'class' should be specified when trying to load study "
                f"named '{self._study_name}' with a `sampler`."
            )

        sampler_class_name = sampler_config.pop("class")
        if sampler_class_name in ["QMCSampler", "CmaEsSampler", "GPSampler"]:
            sampler_config["independent_sampler"] = self._get_sampler(
                sampler_config.pop("independent_sampler")
            )

        if sampler_class_name == "PartialFixedSampler":
            sampler_config["base_sampler"] = self._get_sampler(
                sampler_config.pop("base_sampler")
            )

        sampler_class = getattr(optuna.samplers, sampler_class_name)

        return sampler_class(**sampler_config)

    def _get_pruner(self, pruner_config):
        if pruner_config is None:
            return None

        if "class" not in pruner_config:
            raise ValueError(
                "Optuna `pruner` 'class' should be specified when trying to load study "
                f"named '{self._study_name}' with a `pruner`."
            )

        pruner_class_name = pruner_config.pop("class")
        if pruner_class_name == "PatientPruner":
            pruner_config["wrapped_pruner"] = self._get_pruner(
                pruner_config.pop("wrapped_pruner")
            )

        pruner_class = getattr(optuna.pruners, pruner_class_name)

        return pruner_class(**pruner_config)

    def load(self) -> optuna.Study:
        load_args = deepcopy(self._load_args)
        sampler_config = load_args.pop("sampler")
        sampler = self._get_sampler(sampler_config)

        pruner_config = load_args.pop("pruner")
        pruner = self._get_pruner(pruner_config)

        storage_url_str = self._storage_url.render_as_string(hide_password=False)
        storage = optuna.storages.RDBStorage(url=storage_url_str)
        study = optuna.load_study(
            storage=storage,
            study_name=self._get_load_study_name(),
            sampler=sampler,
            pruner=pruner,
        )

        return study

    def save(self, study: optuna.Study) -> None:
        save_study_name = self._get_save_study_name()

        storage_url_str = self._storage_url.render_as_string(hide_password=False)
        if self._backend == "sqlite":
            os.makedirs(os.path.dirname(self._filepath), exist_ok=True)

            if not os.path.isfile(self._filepath):
                optuna.create_study(
                    storage=storage_url_str,
                )

        storage = optuna.storages.RDBStorage(url=storage_url_str)

        # To overwrite an existing study, we need to first delete it if it exists
        if self._study_name_exists(save_study_name):
            optuna.delete_study(
                storage=storage,
                study_name=save_study_name,
            )

        optuna.copy_study(
            from_study_name=study.study_name,
            from_storage=study._storage,
            to_storage=storage,
            to_study_name=save_study_name,
        )

    def _study_name_exists(self, study_name) -> bool:
        if self._backend == "sqlite" and not os.path.isfile(self._database):
            return False

        storage_url_str = self._storage_url.render_as_string(hide_password=False)
        storage = optuna.storages.RDBStorage(url=storage_url_str)
        study_names = optuna.study.get_all_study_names(storage=storage)
        return study_name in study_names

    def _study_name_glob(self, pattern):
        storage_url_str = self._storage_url.render_as_string(hide_password=False)
        storage = optuna.storages.RDBStorage(url=storage_url_str)
        study_names = optuna.study.get_all_study_names(storage=storage)
        for study_name in study_names:
            if fnmatch.fnmatch(study_name, pattern):
                yield study_name

    def _exists(self) -> bool:
        try:
            load_study_name = self._get_load_study_name()
        except DatasetError:
            return False

        return self._study_name_exists(load_study_name)



"""``APIDataset`` loads the data from HTTP(S) APIs.
It uses the python requests library: https://requests.readthedocs.io/en/latest/
"""
from __future__ import annotations

import json as json_  # make pylint happy
import math
from copy import deepcopy
from typing import Any

import requests
from kedro.io.core import AbstractDataset, DatasetError, parse_dataset_definition
from requests import Session, sessions
from requests.auth import AuthBase

from kedro_datasets.json import JSONDataset
from kedro_datasets.text import TextDataset


class APIDataset(AbstractDataset[None, requests.Response]):
    """``APIDataset`` loads/saves data from/to HTTP(S) APIs.
    It uses the python requests library: https://requests.readthedocs.io/en/latest/

    Examples:
        Using the [YAML API](https://docs.kedro.org/en/stable/catalog-data/data_catalog_yaml_examples/):

        ```yaml
        usda:
          type: api.APIDataset
          url: https://quickstats.nass.usda.gov
          load_args:
              params:
                key: SOME_TOKEN
                format: JSON
                commodity_desc: CORN
                statisticcat_des: YIELD
                agg_level_desc: STATE
                year: 2000
        ```

        Using the [Python API](https://docs.kedro.org/en/stable/catalog-data/advanced_data_catalog_usage/):

        >>> from kedro_datasets.api import APIDataset
        >>>
        >>>
        >>> dataset = APIDataset(
        ...     url="https://api.spaceflightnewsapi.net/v4/articles",
        ...     load_args={
        ...         "params": {
        ...             "news_site": "NASA",
        ...             "launch": "65896761-b6ca-4df3-9699-e077a360c52a",  # Artemis I
        ...         }
        ...     },
        ... )
        >>> data = dataset.load()

        ``APIDataset`` can also be used to save output on a remote server using HTTP(S)
        methods:

        >>> example_table = '{"col1":["val1", "val2"], "col2":["val3", "val4"]}'
        >>>
        >>> dataset = APIDataset(
        ...     method="POST",
        ...     url="https://dummyjson.com/products/add",
        ...     save_args={"chunk_size": 1},
        ... )
        >>> dataset.save(example_table)

        ``APIDataset`` can automatically persist the output of ``POST`` and ``PUT``
        requests via the ``response_dataset`` parameter. This is useful for auditing,
        debugging, or reusing API responses downstream in a pipeline.

        When ``response_dataset`` is configured, the behavior is:

        - For ``JSONDataset``: stores ``response.json()`` (parsed JSON payload)
        - For ``TextDataset``: stores ``response.text`` (raw response body)
        - For other datasets (e.g. ``PickleDataset``, ``MemoryDataset``): stores the
          full ``requests.Response`` object

        You can later retrieve the persisted response by calling
        ``dataset.get_last_response()`` on the dataset instance.

        ```yaml
        api_with_response_storage:
          type: api.APIDataset
          url: https://dummyjson.com/products/add
          method: POST
          response_dataset:
            type: json.JSONDataset
            filepath: data/api_response.json
        ```

        Or using the Python API:

        >>> dataset = APIDataset(
        ...     url="https://dummyjson.com/products/add",
        ...     method="POST",
        ...     response_dataset={"type": "json.JSONDataset", "filepath": "response.json"},
        ... )
        >>> response = dataset.save({"key": "value"})
        >>> # The response data is automatically saved to response.json

    On initialisation, we can specify all the necessary parameters in the save args
    dictionary. The default HTTP(S) method is POST but PUT is also supported. Two
    important parameters to keep in mind are timeout and chunk_size. `timeout` defines
    how long our program waits for a response after a request. `chunk_size`, is only
    used if the input of save method is a list. It will divide the request into chunks
    of size `chunk_size`. For example, here we will send two requests each containing
    one row of our example DataFrame.

    If the data passed to the save method is not a list, ``APIDataset`` will check if it
    can be loaded as JSON. If true, it will send the data unchanged in a single request.
    Otherwise, the ``_save`` method will try to dump the data in JSON format and execute
    the request.
    """

    DEFAULT_SAVE_ARGS = {
        "params": None,
        "headers": None,
        "auth": None,
        "json": None,
        "timeout": 60,
        "chunk_size": 100,
    }

    def __init__(  # noqa: PLR0913
        self,
        *,
        url: str,
        method: str = "GET",
        load_args: dict[str, Any] | None = None,
        save_args: dict[str, Any] | None = None,
        credentials: tuple[str, str] | list[str] | AuthBase | None = None,
        metadata: dict[str, Any] | None = None,
        response_dataset: str | type[AbstractDataset] | dict[str, Any] | None = None,
    ) -> None:
        """Creates a new instance of ``APIDataset`` to fetch data from an API endpoint.

        Args:
            url: The API URL endpoint.
            method: The method of the request. GET, POST, PUT are the only supported
                methods
            load_args: Additional parameters to be fed to requests.request.
                https://requests.readthedocs.io/en/latest/api.html#requests.request
            save_args: Options for saving data on server. Includes all parameters used
                during load method. Adds an optional parameter, ``chunk_size`` which
                determines the size of the package sent at each request.
            credentials: Allows specifying secrets in credentials.yml.
                Expected format is ``('login', 'password')`` if given as a tuple or
                list. An ``AuthBase`` instance can be provided for more complex cases.
            metadata: Any arbitrary metadata.
                This is ignored by Kedro, but may be consumed by users or external plugins.
            response_dataset: Optional dataset to automatically store API responses.
                The API response is stored based on the dataset type:

                - `JSONDataset`: Stores `response.json()` (parsed JSON data)
                - `TextDataset`: Stores `response.text` (response body as string)
                - Other datasets (e.g., `PickleDataset`, `MemoryDataset`): Stores the
                  full `requests.Response` object

                Can be specified as:

                - A string type identifier: `"json.JSONDataset"`
                - A dict with `"type"` key: `{"type": "json.JSONDataset", "filepath": "..."}`
                - A dataset class (advanced usage)

                If `None` (default), responses are not automatically stored.

        Raises:
            ValueError: if both ``auth`` and ``credentials`` are specified or used
                unsupported RESTful API method.
        """
        super().__init__()

        if method == "GET":
            self._params = load_args or {}

        elif method in ["PUT", "POST"]:
            self._params = deepcopy(self.DEFAULT_SAVE_ARGS)
            if save_args is not None:
                self._params.update(save_args)
            self._chunk_size = self._params.pop("chunk_size", 1)
        else:
            raise ValueError("Only GET, POST and PUT methods are supported")

        self._param_auth = self._params.pop("auth", None)

        if credentials is not None and self._param_auth is not None:
            raise ValueError("Cannot specify both auth and credentials.")

        self._auth = credentials or self._param_auth

        if "cert" in self._params:
            self._params["cert"] = self._convert_type(self._params["cert"])

        if "timeout" in self._params:
            self._params["timeout"] = self._convert_type(self._params["timeout"])

        self._request_args: dict[str, Any] = {
            "url": url,
            "method": method,
            "auth": self._convert_type(self._auth),
            **self._params,
        }

        self.metadata = metadata

        # Initialize response dataset if provided
        self._response_dataset_type: type[AbstractDataset[Any, Any]] | None = None
        self._response_dataset_config: dict[str, Any] | None = None
        self._response_dataset_instance: AbstractDataset[Any, Any] | None = None

        if response_dataset is not None:
            dataset_config = (
                response_dataset
                if isinstance(response_dataset, dict)
                else {"type": response_dataset}
            )
            (
                self._response_dataset_type,
                self._response_dataset_config,
            ) = parse_dataset_definition(dataset_config)

    @staticmethod
    def _convert_type(value: Any):
        """
        From the Data Catalog, iterables are provided as Lists.
        However, for some parameters in the Python requests library,
        only Tuples are allowed.
        """
        if isinstance(value, list):
            return tuple(value)
        return value

    @property
    def _response_dataset(self) -> AbstractDataset | None:
        """Lazily create and cache the response dataset instance."""
        if self._response_dataset_type is None:
            return None

        if self._response_dataset_instance is None:
            # Type guard: _response_dataset_config is not None when _response_dataset_type is not None
            assert self._response_dataset_config is not None
            self._response_dataset_instance = self._response_dataset_type(
                **self._response_dataset_config
            )

        return self._response_dataset_instance

    def _describe(self) -> dict[str, Any]:
        # prevent auth from logging
        request_args_cp = self._request_args.copy()
        request_args_cp.pop("auth", None)

        result = dict(request_args_cp)
        if self._response_dataset is not None:
            result["response_dataset"] = self._response_dataset._describe()

        return result

    def _execute_request(self, session: Session) -> requests.Response:
        try:
            response = session.request(**self._request_args)
            response.raise_for_status()
        except requests.exceptions.HTTPError as exc:
            raise DatasetError("Failed to fetch data", exc) from exc
        except OSError as exc:
            raise DatasetError("Failed to connect to the remote server") from exc

        return response

    def get_last_response(self) -> Any:
        if self._response_dataset is None:
            raise DatasetError(
                "No response_dataset configured; cannot retrieve persisted response."
            )

        return self._response_dataset.load()  # type: ignore[return-value]

    def load(self) -> Any:
        if self._request_args["method"] != "GET":
            raise DatasetError(
                "Only GET method is supported for load()."
                "Use save() to send data or get_last_response() to retrieve "
                "a persisted response."
            )

        with sessions.Session() as session:
            return self._execute_request(session)

    def _execute_save_with_chunks(
        self,
        json_data: list[dict[str, Any]],
    ) -> requests.Response:
        chunk_size = self._chunk_size
        n_chunks = math.ceil(len(json_data) / chunk_size)

        for i in range(n_chunks):
            send_data = json_data[i * chunk_size : (i + 1) * chunk_size]
            response = self._execute_save_request(json_data=send_data)

        return response

    def _execute_save_request(self, json_data: Any) -> requests.Response:
        try:
            self._request_args["json"] = json_.loads(json_data)
        except TypeError:
            self._request_args["json"] = json_data
        try:
            response = requests.request(**self._request_args)
            response.raise_for_status()
        except requests.exceptions.HTTPError as exc:
            raise DatasetError("Failed to send data", exc) from exc

        except OSError as exc:
            raise DatasetError("Failed to connect to the remote server") from exc
        return response

    def save(self, data: Any) -> requests.Response:  # type: ignore[override]
        if self._request_args["method"] in ["PUT", "POST"]:
            if isinstance(data, list):
                response: requests.Response = self._execute_save_with_chunks(
                    json_data=data
                )
            else:
                response: requests.Response = self._execute_save_request(json_data=data)

            if self._response_dataset is not None:
                if isinstance(self._response_dataset, JSONDataset):
                    extracted_data = response.json()
                elif isinstance(self._response_dataset, TextDataset):
                    extracted_data = response.text
                else:
                    extracted_data = response

                self._response_dataset.save(extracted_data)

            return response

        raise DatasetError("Use PUT or POST methods for save")

    def _exists(self) -> bool:
        if self._request_args["method"] != "GET":
            return False

        with sessions.Session() as session:
            response = self._execute_request(session)

        return response.ok



import numpy as np
import pytest
from utils import (
    DEFAULT_FOURCC,
    MJPEG_FOURCC,
    MJPEG_FPS,
    MJPEG_LEN,
    MJPEG_SIZE,
    MKV_FOURCC,
    MKV_FPS,
    MKV_LEN,
    MKV_SIZE,
    MP4_FOURCC,
    MP4_FPS,
    MP4_LEN,
    MP4_SIZE,
    TEST_FPS,
    TEST_HEIGHT,
    TEST_NUM_COLOR_FRAMES,
    TEST_WIDTH,
    assert_images_equal,
)

from kedro_datasets_experimental.video.video_dataset import (
    FileVideo,
    GeneratorVideo,
    SequenceVideo,
)


class TestSequenceVideo:
    def test_sequence_video_indexing_first(self, color_video, red_frame):
        """Test indexing a SequenceVideo"""
        red = np.array(color_video[0])
        assert red.shape == (TEST_HEIGHT, TEST_WIDTH, 3)
        assert np.all(red == red_frame)

    def test_sequence_video_indexing_last(self, color_video, purple_frame):
        """Test indexing a SequenceVideo"""
        purple = np.array(color_video[-1])
        assert purple.shape == (TEST_HEIGHT, TEST_WIDTH, 3)
        assert np.all(purple == purple_frame)

    def test_sequence_video_iterable(self, color_video):
        """Test iterating a SequenceVideo"""
        for i, img in enumerate(map(np.array, color_video)):
            assert np.all(img == np.array(color_video[i]))
        assert i == TEST_NUM_COLOR_FRAMES - 1

    def test_sequence_video_fps(self, color_video):
        # Test the one set by the fixture
        assert color_video.fps == TEST_FPS

        # Test creating with another fps
        test_fps_new = 123
        color_video_new = SequenceVideo(color_video._frames, fps=test_fps_new)
        assert color_video_new.fps == test_fps_new

    def test_sequence_video_len(self, color_video):
        assert len(color_video) == TEST_NUM_COLOR_FRAMES

    def test_sequence_video_size(self, color_video):
        assert color_video.size == (TEST_WIDTH, TEST_HEIGHT)

    def test_sequence_video_fourcc_default_value(self, color_video):
        assert color_video.fourcc == DEFAULT_FOURCC

    def test_sequence_video_fourcc(self, color_video):
        fourcc_new = "mjpg"
        assert (
            DEFAULT_FOURCC != fourcc_new
        ), "Test does not work if new test value is same as default"
        color_video_new = SequenceVideo(
            color_video._frames, fps=TEST_FPS, fourcc=fourcc_new
        )
        assert color_video_new.fourcc == fourcc_new


class TestGeneratorVideo:
    def test_generator_video_iterable(self, color_video_generator, color_video):
        """Test iterating a GeneratorVideo

        The content of the mock GeneratorVideo should be the same as the SequenceVideo,
        the content in the later is tested in other unit tests and can thus be trusted
        """
        for i, img in enumerate(map(np.array, color_video_generator)):
            assert np.all(img == np.array(color_video[i]))
        assert i == TEST_NUM_COLOR_FRAMES - 1

    def test_generator_video_fps(self, color_video_generator):
        # Test the one set by the fixture
        assert color_video_generator.fps == TEST_FPS

        # Test creating with another fps
        test_fps_new = 123
        color_video_new = GeneratorVideo(
            color_video_generator._gen, length=TEST_NUM_COLOR_FRAMES, fps=test_fps_new
        )
        assert color_video_new.fps == test_fps_new

    def test_generator_video_len(self, color_video_generator):
        assert len(color_video_generator) == TEST_NUM_COLOR_FRAMES

    def test_generator_video_size(self, color_video_generator):
        assert color_video_generator.size == (TEST_WIDTH, TEST_HEIGHT)

    def test_generator_video_fourcc_default_value(self, color_video_generator):
        assert color_video_generator.fourcc == DEFAULT_FOURCC

    def test_generator_video_fourcc(self, color_video_generator):
        fourcc_new = "mjpg"
        assert (
            DEFAULT_FOURCC != fourcc_new
        ), "Test does not work if new test value is same as default"
        color_video_new = GeneratorVideo(
            color_video_generator._gen,
            length=TEST_NUM_COLOR_FRAMES,
            fps=TEST_FPS,
            fourcc=fourcc_new,
        )
        assert color_video_new.fourcc == fourcc_new


class TestFileVideo:
    @pytest.mark.skip(reason="Can't deal with videos with missing time info")
    def test_file_props_mjpeg(self, mjpeg_object):
        assert mjpeg_object.fourcc == MJPEG_FOURCC
        assert mjpeg_object.fps == MJPEG_FPS
        assert mjpeg_object.size == MJPEG_SIZE
        assert len(mjpeg_object) == MJPEG_LEN

    def test_file_props_mkv(self, mkv_object):
        assert mkv_object.fourcc == MKV_FOURCC
        assert mkv_object.fps == MKV_FPS
        assert mkv_object.size == MKV_SIZE
        assert len(mkv_object) == MKV_LEN

    @pytest.mark.skip(reason="Skipping for now, need to investigate")
    def test_file_props_mp4(self, mp4_object):
        assert mp4_object.fourcc == MP4_FOURCC
        assert mp4_object.fps == MP4_FPS
        assert mp4_object.size == MP4_SIZE
        assert len(mp4_object) == MP4_LEN

    def test_file_index_first(self, color_video_object, red_frame):
        assert_images_equal(color_video_object[0], red_frame)

    def test_file_index_last_by_index(self, color_video_object, purple_frame):
        assert_images_equal(color_video_object[TEST_NUM_COLOR_FRAMES - 1], purple_frame)

    def test_file_index_last(self, color_video_object, purple_frame):
        assert_images_equal(color_video_object[-1], purple_frame)

    def test_file_video_failed_capture(self, mocker):
        """Validate good behavior on failed decode

        The best behavior in this case is not obvious, the len property of the
        video object specifies more frames than is actually possible to decode. We
        cannot know this in advance without spending loads of time to decode all frames
        in order to count them."""
        mock_cv2 = mocker.patch("kedro_datasets_experimental.video.video_dataset.cv2")
        mock_cap = mock_cv2.VideoCapture.return_value = mocker.Mock()
        mock_cap.get.return_value = 2  # Set the length of the video
        ds = FileVideo("/a/b/c")

        mock_cap.read.return_value = True, np.zeros((1, 1))
        assert ds[0]

        mock_cap.read.return_value = False, None
        with pytest.raises(IndexError):
            ds[1]



"""``APIDataset`` loads the data from HTTP(S) APIs
and returns them into either as string or json Dict.
It uses the python requests library: https://requests.readthedocs.io/en/latest/
"""

from typing import Any

import lazy_loader as lazy

try:
    from .api_dataset import APIDataset
except (ImportError, RuntimeError):
    # For documentation builds that might fail due to dependency issues
    # https://github.com/pylint-dev/pylint/issues/4300#issuecomment-1043601901
    APIDataset: Any

__getattr__, __dir__, __all__ = lazy.attach(
    __name__, submod_attrs={"api_dataset": ["APIDataset"]}
)



"""``PDFDataset`` loads data from PDF files using an underlying
filesystem (e.g.: local, S3, GCS). It uses pypdf to read and extract text from PDF files.
"""
from __future__ import annotations

from copy import deepcopy
from pathlib import PurePosixPath
from typing import Any, NoReturn

import fsspec
import pypdf
from kedro.io.core import (
    AbstractDataset,
    DatasetError,
    get_filepath_str,
    get_protocol_and_path,
)


class PDFDataset(AbstractDataset[NoReturn, list[str]]):
    """``PDFDataset`` loads data from PDF files using an underlying
    filesystem (e.g.: local, S3, GCS). It uses pypdf to read and extract text from PDF files.

    This is a read-only dataset - saving is not supported.

    Examples:
        Using the [YAML API](https://docs.kedro.org/en/stable/catalog-data/data_catalog_yaml_examples/):

        ```yaml
        my_pdf_document:
          type: pypdf.PDFDataset
          filepath: data/01_raw/document.pdf

        password_protected_pdf:
          type: pypdf.PDFDataset
          filepath: data/01_raw/protected.pdf
          load_args:
            password: "pass123"  # pragma: allowlist secret

        s3_pdf:
          type: pypdf.PDFDataset
          filepath: s3://your_bucket/document.pdf
          credentials: dev_s3
        ```

        Using the [Python API](https://docs.kedro.org/en/stable/catalog-data/advanced_data_catalog_usage/):

        >>> from kedro_datasets_experimental.pypdf import PDFDataset
        >>>
        >>> dataset = PDFDataset(filepath="data/document.pdf")
        >>> pages = dataset.load()
        >>> # pages is a list of strings, one per page
        >>> assert isinstance(pages, list)
        >>> assert all(isinstance(page, str) for page in pages)

    """

    DEFAULT_LOAD_ARGS: dict[str, Any] = {"strict": False}

    def __init__(
        self,
        *,
        filepath: str,
        load_args: dict[str, Any] | None = None,
        credentials: dict[str, Any] | None = None,
        fs_args: dict[str, Any] | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        """Creates a new instance of ``PDFDataset`` pointing to a concrete PDF file
        on a specific filesystem.

        Args:
            filepath: Filepath in POSIX format to a PDF file prefixed with a protocol like `s3://`.
                If prefix is not provided, `file` protocol (local filesystem) will be used.
                The prefix should be any protocol supported by ``fsspec``.
            load_args: Pypdf options for loading PDF files (arguments passed
                into ``pypdf.PdfReader``). Here you can find all available arguments:
                https://pypdf.readthedocs.io/en/stable/modules/PdfReader.html
                All defaults are preserved, except "strict", which is set to False.
                Common options include:
                - password (str): Password for encrypted PDFs
                - strict (bool): Whether to raise errors on malformed PDFs (default: False)
            credentials: Credentials required to get access to the underlying filesystem.
                E.g. for ``GCSFileSystem`` it should look like `{"token": None}`.
            fs_args: Extra arguments to pass into underlying filesystem class constructor
                (e.g. `{"project": "my-project"}` for ``GCSFileSystem``), as well as
                to pass to the filesystem's `open` method through nested keys
                `open_args_load` and `open_args_save`.
                Here you can find all available arguments for `open`:
                https://filesystem-spec.readthedocs.io/en/latest/api.html#fsspec.spec.AbstractFileSystem.open
                All defaults are preserved.
            metadata: Any arbitrary metadata.
                This is ignored by Kedro, but may be consumed by users or external plugins.
        """
        _fs_args = deepcopy(fs_args) or {}
        _fs_open_args_load = _fs_args.pop("open_args_load", {})
        _credentials = deepcopy(credentials) or {}

        super().__init__()

        protocol, path = get_protocol_and_path(filepath)
        if protocol == "file":
            _fs_args.setdefault("auto_mkdir", True)

        self._protocol = protocol
        self._fs = fsspec.filesystem(self._protocol, **_credentials, **_fs_args)
        self._filepath = PurePosixPath(path)
        self.metadata = metadata

        # Handle default load and fs arguments
        self._load_args = {**self.DEFAULT_LOAD_ARGS, **(load_args or {})}
        self._fs_open_args_load = _fs_open_args_load or {}

    def _describe(self) -> dict[str, Any]:
        return {
            "filepath": self._filepath,
            "protocol": self._protocol,
            "load_args": self._load_args,
        }

    def load(self) -> list[str]:
        """Loads data from a PDF file.

        Returns:
            list[str]: A list of strings, where each string contains the text extracted from one page.
        """
        load_path = get_filepath_str(self._filepath, self._protocol)

        with self._fs.open(load_path, mode="rb", **self._fs_open_args_load) as fs_file:
            pdf_reader = pypdf.PdfReader(stream=fs_file, **self._load_args)
            pages = []
            for page in pdf_reader.pages:
                pages.append(page.extract_text())
            return pages

    def save(self, data: NoReturn) -> None:
        """Saving to PDFDataset is not supported.

        Args:
            data: Data to save.

        Raises:
            DatasetError: Always raised as saving is not supported.
        """
        raise DatasetError("Saving to PDFDataset is not supported.")

    def _exists(self) -> bool:
        """Check if the PDF file exists.

        Returns:
            bool: True if the file exists, False otherwise.
        """
        load_path = get_filepath_str(self._filepath, self._protocol)
        return self._fs.exists(load_path)

    def _release(self) -> None:
        """Release any cached filesystem information."""
        self._invalidate_cache()

    def _invalidate_cache(self) -> None:
        """Invalidate underlying filesystem caches."""
        filepath = get_filepath_str(self._filepath, self._protocol)
        self._fs.invalidate_cache(filepath)



"""``CSVDataset`` loads/saves data from/to a CSV file using an underlying
filesystem (e.g.: local, S3, GCS). It uses polars to handle the CSV file.
"""
from __future__ import annotations

import logging
from copy import deepcopy
from pathlib import PurePosixPath
from typing import Any

import fsspec
import polars as pl
from kedro.io.core import (
    PROTOCOL_DELIMITER,
    AbstractVersionedDataset,
    DatasetError,
    Version,
    get_filepath_str,
    get_protocol_and_path,
)

from kedro_datasets._typing import TablePreview

logger = logging.getLogger(__name__)


class CSVDataset(AbstractVersionedDataset[pl.DataFrame, pl.DataFrame]):
    """``CSVDataset`` loads/saves data from/to a CSV file using an underlying
    filesystem (e.g.: local, S3, GCS). It uses polars to handle the CSV file.

    Examples:
        Using the [YAML API](https://docs.kedro.org/en/stable/catalog-data/data_catalog_yaml_examples/):

        ```yaml
        cars:
          type: polars.CSVDataset
          filepath: data/01_raw/company/cars.csv
          load_args:
            sep: ","
            parse_dates: False
          save_args:
            has_header: False
            null_value: "somenullstring"

        motorbikes:
          type: polars.CSVDataset
          filepath: s3://your_bucket/data/02_intermediate/company/motorbikes.csv
          credentials: dev_s3
        ```

        Using the [Python API](https://docs.kedro.org/en/stable/catalog-data/advanced_data_catalog_usage/):

        >>> import sys
        >>>
        >>> import polars as pl
        >>> import pytest
        >>> from kedro_datasets.polars import CSVDataset
        >>>
        >>> if sys.platform.startswith("win"):
        ...     pytest.skip("this doctest fails on Windows CI runner")
        ...
        >>> data = pl.DataFrame({"col1": [1, 2], "col2": [4, 5], "col3": [5, 6]})
        >>>
        >>> dataset = CSVDataset(filepath=tmp_path / "test.csv")
        >>> dataset.save(data)
        >>> reloaded = dataset.load()
        >>> assert data.equals(reloaded)

    """

    DEFAULT_LOAD_ARGS: dict[str, Any] = {"rechunk": True}
    DEFAULT_SAVE_ARGS: dict[str, Any] = {}
    DEFAULT_FS_ARGS: dict[str, Any] = {
        "open_args_save": {"mode": "w", "encoding": "utf-8"}
    }

    def __init__(  # noqa: PLR0913
        self,
        *,
        filepath: str,
        load_args: dict[str, Any] | None = None,
        save_args: dict[str, Any] | None = None,
        version: Version | None = None,
        credentials: dict[str, Any] | None = None,
        fs_args: dict[str, Any] | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        """Creates a new instance of ``CSVDataset`` pointing to a concrete CSV file
        on a specific filesystem.

        Args:
            filepath: Filepath in POSIX format to a CSV file prefixed with a protocol
                `s3://`.
                If prefix is not provided, `file` protocol (local filesystem)
                will be used.
                The prefix should be any protocol supported by ``fsspec``.
                Note: `http(s)` doesn't support versioning.
            load_args: Polars options for loading CSV files.
                Here you can find all available arguments:
                https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.read_csv.html#polars.read_csv
                All defaults are preserved, but we explicitly use `rechunk=True` for `seaborn`
                compatibility.
            save_args: Polars options for saving CSV files.
                Here you can find all available arguments:
                https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.DataFrame.write_csv.html
                All defaults are preserved.
            version: If specified, should be an instance of
                ``kedro.io.core.Version``. If its ``load`` attribute is
                None, the latest version will be loaded. If its ``save``
                attribute is None, save version will be autogenerated.
            credentials: Credentials required to get access to the underlying filesystem.
                E.g. for ``GCSFileSystem`` it should look like `{"token": None}`.
            fs_args: Extra arguments to pass into underlying filesystem class constructor
                (e.g. `{"project": "my-project"}` for ``GCSFileSystem``).
                Defaults are preserved, apart from the `open_args_save` `mode` which is set to `w`.

            metadata: Any arbitrary metadata.
                This is ignored by Kedro, but may be consumed by users or external plugins.
        """
        _fs_args = deepcopy(fs_args) or {}
        _fs_open_args_load = _fs_args.pop("open_args_load", {})
        _fs_open_args_save = _fs_args.pop("open_args_save", {})
        _credentials = deepcopy(credentials) or {}

        protocol, path = get_protocol_and_path(filepath, version)
        if protocol == "file":
            _fs_args.setdefault("auto_mkdir", True)

        self._protocol = protocol
        self._storage_options = {**_credentials, **_fs_args}
        self._fs = fsspec.filesystem(self._protocol, **self._storage_options)

        self.metadata = metadata

        super().__init__(
            filepath=PurePosixPath(path),
            version=version,
            exists_function=self._fs.exists,
            glob_function=self._fs.glob,
        )

        # Handle default load and save and fs arguments
        self._load_args = {**self.DEFAULT_LOAD_ARGS, **(load_args or {})}
        self._save_args = {**self.DEFAULT_SAVE_ARGS, **(save_args or {})}
        self._fs_open_args_load = {
            **self.DEFAULT_FS_ARGS.get("open_args_load", {}),
            **(_fs_open_args_load or {}),
        }
        self._fs_open_args_save = {
            **self.DEFAULT_FS_ARGS.get("open_args_save", {}),
            **(_fs_open_args_save or {}),
        }

        if "storage_options" in self._save_args or "storage_options" in self._load_args:
            logger.warning(
                "Dropping 'storage_options' for %s, "
                "please specify them under 'fs_args' or 'credentials'.",
                self._filepath,
            )
            self._save_args.pop("storage_options", None)
            self._load_args.pop("storage_options", None)

    def _describe(self) -> dict[str, Any]:
        return {
            "filepath": self._filepath,
            "protocol": self._protocol,
            "load_args": self._load_args,
            "save_args": self._save_args,
            "version": self._version,
        }

    def load(self) -> pl.DataFrame:
        load_path = str(self._get_load_path())
        if self._protocol == "file":
            # file:// protocol seems to misbehave on Windows
            # (<urlopen error file not on local host>),
            # so we don't join that back to the filepath;
            # storage_options also don't work with local paths
            return pl.read_csv(load_path, **self._load_args)

        load_path = f"{self._protocol}{PROTOCOL_DELIMITER}{load_path}"
        return pl.read_csv(
            load_path, storage_options=self._storage_options, **self._load_args
        )

    def save(self, data: pl.DataFrame) -> None:
        save_path = get_filepath_str(self._get_save_path(), self._protocol)

        with self._fs.open(save_path, **self._fs_open_args_save) as fs_file:
            data.write_csv(file=fs_file, **self._save_args)

        self._invalidate_cache()

    def _exists(self) -> bool:
        try:
            load_path = get_filepath_str(self._get_load_path(), self._protocol)
        except DatasetError:
            return False

        return self._fs.exists(load_path)

    def _release(self) -> None:
        super()._release()
        self._invalidate_cache()

    def _invalidate_cache(self) -> None:
        """Invalidate underlying filesystem caches."""
        filepath = get_filepath_str(self._filepath, self._protocol)
        self._fs.invalidate_cache(filepath)

    def preview(self, nrows: int = 5) -> TablePreview:
        """
        Generate a preview of the dataset with a specified number of rows.

        Args:
            nrows: The number of rows to include in the preview. Defaults to 5.

        Returns:
            dict: A dictionary containing the data in a split format.
        """
        # Create a copy so it doesn't contaminate the original dataset
        dataset_copy = self._copy()
        data = dataset_copy.load().limit(nrows if type(nrows) is int else 5)
        data_dict = data.to_pandas().to_dict(orient="split")
        return TablePreview(data_dict)



import pytest
import torch
import torch.nn.functional as F
from torch import nn

from kedro_datasets_experimental.pytorch import PyTorchDataset


# Define model
class TheModelClass(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = x.view(-1, 16 * 5 * 5)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


@pytest.fixture
def filepath_model(tmp_path):
    return (tmp_path / "model.pt").as_posix()


@pytest.fixture
def pytorch_dataset(filepath_model, load_args, save_args, fs_args):
    return PyTorchDataset(
        filepath=filepath_model,
        load_args=load_args,
        save_args=save_args,
        fs_args=fs_args,
    )


@pytest.fixture
def versioned_pytorch_dataset(filepath_model, load_version, save_version):
    return PyTorchDataset(
        filepath=filepath_model, load_version=load_version, save_version=save_version
    )


@pytest.fixture
def dummy_model():
    return TheModelClass()


class TestPyTorchDataset:
    def test_save_and_load_dataset(self, pytorch_dataset, dummy_model, filepath_model):
        pytorch_dataset.save(dummy_model)
        model = TheModelClass()
        model.load_state_dict(pytorch_dataset.load())
        reloaded_state_dict = model.state_dict()
        dummy_state_dict = dummy_model.state_dict()
        for key, value in reloaded_state_dict.items():
            assert torch.equal(dummy_state_dict[key], value)



"""``AbstractDataset`` implementation to load/save data from/to OpenXML files."""

from typing import Any

import lazy_loader as lazy

try:
    from .docx_dataset import DocxDataset
except (ImportError, RuntimeError):
    DocxDataset: Any

try:
    from .pptx_dataset import PptxDataset
except (ImportError, RuntimeError):
    PptxDataset: Any

# https://github.com/pylint-dev/pylint/issues/4300#issuecomment-1043601901

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={"docx_dataset": ["DocxDataset"], "pptx_dataset": ["PptxDataset"]},
)



"""``AbstractDataset`` implementation to load/save torch models using PyTorch's built-in methods """

from typing import Any

import lazy_loader as lazy

try:
    from .pytorch_dataset import PyTorchDataset
except (ImportError, RuntimeError):
    # For documentation builds that might fail due to dependency issues
    # https://github.com/pylint-dev/pylint/issues/4300#issuecomment-1043601901
    PyTorchDataset: Any

__getattr__, __dir__, __all__ = lazy.attach(
    __name__, submod_attrs={"pytorch_dataset": ["PyTorchDataset"]}
)



from pathlib import Path

import pytest
from PIL import Image
from utils import TEST_FPS, TEST_HEIGHT, TEST_WIDTH

from kedro_datasets_experimental.video.video_dataset import (
    FileVideo,
    GeneratorVideo,
    SequenceVideo,
)


@pytest.fixture(scope="module")
def red_frame():
    return Image.new("RGB", (TEST_WIDTH, TEST_HEIGHT), (255, 0, 0))


@pytest.fixture(scope="module")
def green_frame():
    return Image.new("RGB", (TEST_WIDTH, TEST_HEIGHT), (0, 255, 0))


@pytest.fixture(scope="module")
def blue_frame():
    return Image.new("RGB", (TEST_WIDTH, TEST_HEIGHT), (0, 0, 255))


@pytest.fixture(scope="module")
def yellow_frame():
    return Image.new("RGB", (TEST_WIDTH, TEST_HEIGHT), (255, 255, 0))


@pytest.fixture(scope="module")
def purple_frame():
    return Image.new("RGB", (TEST_WIDTH, TEST_HEIGHT), (255, 0, 255))


@pytest.fixture
def color_video(red_frame, green_frame, blue_frame, yellow_frame, purple_frame):
    return SequenceVideo(
        [red_frame, green_frame, blue_frame, yellow_frame, purple_frame],
        fps=TEST_FPS,
    )


@pytest.fixture
def color_video_generator(
    red_frame, green_frame, blue_frame, yellow_frame, purple_frame
):
    sequence = [red_frame, green_frame, blue_frame, yellow_frame, purple_frame]

    def generator():
        yield from sequence

    return GeneratorVideo(
        generator(),
        length=len(sequence),
        fps=TEST_FPS,
    )


@pytest.fixture
def filepath_mp4():
    """This is a real video converted to mp4/h264 with ffmpeg command"""
    return str(Path(__file__).parent / "data/video.mp4")


@pytest.fixture
def filepath_mkv():
    """This a a real video recoreded with an Axis network camera"""
    return str(Path(__file__).parent / "data/video.mkv")


@pytest.fixture
def filepath_mjpeg():
    """This is a real video recorded with an Axis network camera"""
    return str(Path(__file__).parent / "data/video.mjpeg")


@pytest.fixture
def filepath_color_mp4():
    """This is a video created with the OpenCV VideoWriter

    it contains 5 frames which each is a single color: red, green, blue, yellow, purple
    """
    return str(Path(__file__).parent / "data/color_video.mp4")


@pytest.fixture
def mp4_object(filepath_mp4):
    return FileVideo(filepath_mp4)


@pytest.fixture
def mkv_object(filepath_mkv):
    return FileVideo(filepath_mkv)


@pytest.fixture
def mjpeg_object(filepath_mjpeg):
    return FileVideo(filepath_mjpeg)


@pytest.fixture
def color_video_object(filepath_color_mp4):
    return FileVideo(filepath_color_mp4)



"""Provide data loading and saving functionality for Optuna's study."""

from typing import Any

import lazy_loader as lazy

StudyDataset: Any

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={
        "study_dataset": ["StudyDataset"],
    },
)



"""``AbstractDataset`` implementation to save matplotlib objects as image files."""

from typing import Any

import lazy_loader as lazy

try:
    from .matplotlib_dataset import MatplotlibDataset
except (ImportError, RuntimeError):
    # For documentation builds that might fail due to dependency issues
    # https://github.com/pylint-dev/pylint/issues/4300#issuecomment-1043601901
    MatplotlibDataset: Any

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={
        "matplotlib_dataset": ["MatplotlibDataset"],
    },
)



"""``ImageDataset`` loads/saves image data as `numpy` from an underlying
filesystem (e.g.: local, S3, GCS). It uses Pillow to handle image file.
"""

from __future__ import annotations

from copy import deepcopy
from pathlib import PurePosixPath
from typing import Any

import fsspec
from kedro.io.core import (
    AbstractVersionedDataset,
    DatasetError,
    Version,
    get_filepath_str,
    get_protocol_and_path,
)
from PIL import Image


class ImageDataset(AbstractVersionedDataset[Image.Image, Image.Image]):
    """``ImageDataset`` loads/saves image data as `numpy` from an underlying
    filesystem (e.g.: local, S3, GCS). It uses Pillow to handle image file.

    Examples:
        Using the [Python API](https://docs.kedro.org/en/stable/catalog-data/advanced_data_catalog_usage/):

        >>> from PIL import Image
        >>> import tempfile
        >>> from kedro_datasets.pillow import ImageDataset
        >>>
        >>> tmp = tempfile.NamedTemporaryFile(suffix=".jpg", delete=False)
        >>> img = Image.new("RGB", (10, 10), color="red")
        >>> img.save(tmp.name)
        >>> tmp.close()
        >>>
        >>> dataset = ImageDataset(filepath=tmp.name)
        >>> loaded_img = dataset.load()
        >>> loaded_img.size
        (10, 10)

    """

    DEFAULT_SAVE_ARGS: dict[str, Any] = {}
    DEFAULT_FS_ARGS: dict[str, Any] = {"open_args_save": {"mode": "wb"}}

    def __init__(  # noqa: PLR0913
        self,
        *,
        filepath: str,
        save_args: dict[str, Any] | None = None,
        version: Version | None = None,
        credentials: dict[str, Any] | None = None,
        fs_args: dict[str, Any] | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        """Creates a new instance of ``ImageDataset`` pointing to a concrete image file
        on a specific filesystem.

        Args:
            filepath: Filepath in POSIX format to an image file prefixed with a protocol like
                `s3://`. If prefix is not provided, `file` protocol (local filesystem) will be used.
                The prefix should be any protocol supported by ``fsspec``.
                Note: `http(s)` doesn't support versioning.
            save_args: Pillow options for saving image files.
                Here you can find all available arguments:
                https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.save
                All defaults are preserved.
            version: If specified, should be an instance of
                ``kedro.io.core.Version``. If its ``load`` attribute is
                None, the latest version will be loaded. If its ``save``
                attribute is None, save version will be autogenerated.
            credentials: Credentials required to get access to the underlying filesystem.
                E.g. for ``GCSFileSystem`` it should look like `{"token": None}`.
            fs_args: Extra arguments to pass into underlying filesystem class constructor
                (e.g. `{"project": "my-project"}` for ``GCSFileSystem``), as well as
                to pass to the filesystem's `open` method through nested keys
                `open_args_load` and `open_args_save`.
                Here you can find all available arguments for `open`:
                https://filesystem-spec.readthedocs.io/en/latest/api.html#fsspec.spec.AbstractFileSystem.open
                All defaults are preserved, except `mode`, which is set to `wb` when saving.
            metadata: Any arbitrary metadata.
                This is ignored by Kedro, but may be consumed by users or external plugins.
        """
        _fs_args = deepcopy(fs_args) or {}
        _fs_open_args_load = _fs_args.pop("open_args_load", {})
        _fs_open_args_save = _fs_args.pop("open_args_save", {})
        _credentials = deepcopy(credentials) or {}

        protocol, path = get_protocol_and_path(filepath, version)
        if protocol == "file":
            _fs_args.setdefault("auto_mkdir", True)

        self._protocol = protocol
        self._fs = fsspec.filesystem(self._protocol, **_credentials, **_fs_args)

        self.metadata = metadata

        super().__init__(
            filepath=PurePosixPath(path),
            version=version,
            exists_function=self._fs.exists,
            glob_function=self._fs.glob,
        )

        # Handle default save and fs arguments
        self._save_args = {**self.DEFAULT_SAVE_ARGS, **(save_args or {})}
        self._fs_open_args_load = {
            **self.DEFAULT_FS_ARGS.get("open_args_load", {}),
            **(_fs_open_args_load or {}),
        }
        self._fs_open_args_save = {
            **self.DEFAULT_FS_ARGS.get("open_args_save", {}),
            **(_fs_open_args_save or {}),
        }

    def _describe(self) -> dict[str, Any]:
        return {
            "filepath": self._filepath,
            "protocol": self._protocol,
            "save_args": self._save_args,
            "version": self._version,
        }

    def load(self) -> Image.Image:
        load_path = get_filepath_str(self._get_load_path(), self._protocol)

        with self._fs.open(load_path, **self._fs_open_args_load) as fs_file:
            return Image.open(fs_file).copy()

    def save(self, data: Image.Image) -> None:
        save_path = self._get_save_path()

        with self._fs.open(
            get_filepath_str(save_path, self._protocol), **self._fs_open_args_save
        ) as fs_file:
            data.save(fs_file, format=self._get_format(save_path), **self._save_args)

        self._invalidate_cache()

    @staticmethod
    def _get_format(file_path: PurePosixPath):
        ext = file_path.suffix.lower()
        if ext not in Image.EXTENSION:
            Image.init()
        return Image.EXTENSION.get(ext)

    def _exists(self) -> bool:
        try:
            load_path = get_filepath_str(self._get_load_path(), self._protocol)
        except DatasetError:
            return False

        return self._fs.exists(load_path)

    def _release(self) -> None:
        super()._release()
        self._invalidate_cache()

    def _invalidate_cache(self) -> None:
        """Invalidate underlying filesystem caches."""
        filepath = get_filepath_str(self._filepath, self._protocol)
        self._fs.invalidate_cache(filepath)
